NAME = "PyChess"

# Player types
LOCAL, ARTIFICIAL, REMOTE = range(3)

# Engine strengths
EASY, INTERMEDIATE, EXPERT = range(3)

# Tools
TOOL_NONE, TOOL_CHESSDB, TOOL_SCOUTFISH = range(3)

# Player colors
WHITE, BLACK = range(2)

# Game states
(
    WAITING_TO_START,
    PAUSED,
    RUNNING,
    DRAW,
    WHITEWON,
    BLACKWON,
    KILLED,
    ADJOURNED,
    ABORTED,
    UNKNOWN_STATE,
    ICC_ABORTED,
    CANCELLED,
) = range(12)
reprResult = ["*", "*", "*", "1/2-1/2", "1-0", "0-1", "*", "*", "*", "*", "*", "*"]

UNDOABLE_STATES = (DRAW, WHITEWON, BLACKWON)
UNFINISHED_STATES = (WAITING_TO_START, PAUSED, RUNNING, UNKNOWN_STATE)

# Chess variants
(
    NORMALCHESS,
    CORNERCHESS,
    SHUFFLECHESS,
    FISCHERRANDOMCHESS,
    RANDOMCHESS,
    ASYMMETRICRANDOMCHESS,
    UPSIDEDOWNCHESS,
    PAWNSPUSHEDCHESS,
    PAWNSPASSEDCHESS,
    THEBANCHESS,
    PAWNODDSCHESS,
    KNIGHTODDSCHESS,
    ROOKODDSCHESS,
    QUEENODDSCHESS,
    ATOMICCHESS,
    BUGHOUSECHESS,
    CRAZYHOUSECHESS,
    LOSERSCHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    WILDCASTLECHESS,
    WILDCASTLESHUFFLECHESS,
    KINGOFTHEHILLCHESS,
    THREECHECKCHESS,
    HORDECHESS,
    RACINGKINGSCHESS,
    ASEANCHESS,
    MAKRUKCHESS,
    SITTUYINCHESS,
    CAMBODIANCHESS,
    AIWOKCHESS,
    EUROSHOGICHESS,
    SETUPCHESS,
    PLACEMENTCHESS,
    SCHESS,
    LIGHTBRIGADECHESS,
) = range(36)

ASEAN_VARIANTS = (ASEANCHESS, MAKRUKCHESS, CAMBODIANCHESS, AIWOKCHESS, SITTUYINCHESS)
DROP_VARIANTS = (
    BUGHOUSECHESS,
    CRAZYHOUSECHESS,
    EUROSHOGICHESS,
    SITTUYINCHESS,
    SETUPCHESS,
    PLACEMENTCHESS,
    SCHESS,
)
UNSUPPORTED = (BUGHOUSECHESS, AIWOKCHESS, EUROSHOGICHESS, SETUPCHESS)

# Chess variant groups
(
    VARIANTS_ODDS,
    VARIANTS_SHUFFLE,
    VARIANTS_OTHER,
    VARIANTS_OTHER_NONSTANDARD,
    VARIANTS_ASEAN,
) = range(5)

# Action errors
(
    ACTION_ERROR_NOT_OUT_OF_TIME,
    ACTION_ERROR_CLOCK_NOT_STARTED,
    ACTION_ERROR_SWITCH_UNDERWAY,
    ACTION_ERROR_CLOCK_NOT_PAUSED,
    ACTION_ERROR_TOO_LARGE_UNDO,
    ACTION_ERROR_NONE_TO_ACCEPT,
    ACTION_ERROR_NONE_TO_WITHDRAW,
    ACTION_ERROR_NONE_TO_DECLINE,
) = range(8)

# Game state reasons
(
    ABORTED_ADJUDICATION,
    ABORTED_AGREEMENT,
    ABORTED_COURTESY,
    ABORTED_EARLY,
    ABORTED_SERVER_SHUTDOWN,
    ADJOURNED_COURTESY,
    ABORTED_DISCONNECTION,
    ADJOURNED_AGREEMENT,
    ADJOURNED_LOST_CONNECTION,
    ADJOURNED_SERVER_SHUTDOWN,
    ADJOURNED_COURTESY_WHITE,
    ADJOURNED_COURTESY_BLACK,
    ADJOURNED_LOST_CONNECTION_WHITE,
    ADJOURNED_LOST_CONNECTION_BLACK,
    DRAW_50MOVES,
    DRAW_ADJUDICATION,
    DRAW_AGREE,
    DRAW_CALLFLAG,
    DRAW_INSUFFICIENT,
    DRAW_EQUALMATERIAL,
    DRAW_LENGTH,
    DRAW_REPETITION,
    DRAW_STALEMATE,
    DRAW_KINGSINEIGHTROW,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    DRAW_WHITEINSUFFICIENTANDBLACKTIME,
    WON_ADJUDICATION,
    WON_CALLFLAG,
    WON_DISCONNECTION,
    WON_MATE,
    WON_RESIGN,
    WON_LESSMATERIAL,
    WON_NOMATERIAL,
    WON_KINGEXPLODE,
    WON_KINGINCENTER,
    WON_THREECHECK,
    WON_KINGINEIGHTROW,
    WON_WIPEOUT,
    PRACTICE_GOAL_REACHED,
    WHITE_ENGINE_DIED,
    BLACK_ENGINE_DIED,
    DISCONNECTED,
    UNKNOWN_REASON,
) = range(43)

ABORTED_REASONS = (
    ABORTED_ADJUDICATION,
    ABORTED_AGREEMENT,
    ABORTED_COURTESY,
    ABORTED_EARLY,
    ABORTED_SERVER_SHUTDOWN,
    ABORTED_DISCONNECTION,
)

ADJOURNED_REASONS = (
    ADJOURNED_COURTESY,
    ADJOURNED_AGREEMENT,
    ADJOURNED_LOST_CONNECTION,
    ADJOURNED_SERVER_SHUTDOWN,
    ADJOURNED_COURTESY_WHITE,
    ADJOURNED_COURTESY_BLACK,
    ADJOURNED_LOST_CONNECTION_WHITE,
    ADJOURNED_LOST_CONNECTION_BLACK,
)

UNDOABLE_REASONS = (
    DRAW_50MOVES,
    DRAW_INSUFFICIENT,
    DRAW_EQUALMATERIAL,
    DRAW_REPETITION,
    DRAW_STALEMATE,
    DRAW_AGREE,
    DRAW_CALLFLAG,
    DRAW_KINGSINEIGHTROW,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    DRAW_WHITEINSUFFICIENTANDBLACKTIME,
    WON_MATE,
    WON_NOMATERIAL,
    WON_CALLFLAG,
    WON_RESIGN,
    WON_LESSMATERIAL,
    WON_KINGEXPLODE,
    WON_KINGINCENTER,
    WON_THREECHECK,
    WON_KINGINEIGHTROW,
    WON_WIPEOUT,
)

ADJUDICATION_REASONS = (DRAW_ADJUDICATION, WON_ADJUDICATION)

DEATH_REASONS = (WHITE_ENGINE_DIED, BLACK_ENGINE_DIED, DISCONNECTED)

CALLFLAG_REASONS = (DRAW_CALLFLAG, WON_CALLFLAG)

# Player actions
RESIGNATION = "resignation"
FLAG_CALL = "flag call"
DRAW_OFFER = "draw offer"
ABORT_OFFER = "abort offer"
ADJOURN_OFFER = "adjourn offer"
PAUSE_OFFER = "pause offer"
RESUME_OFFER = "resume offer"
SWITCH_OFFER = "switch offer"
TAKEBACK_OFFER = "takeback offer"
MATCH_OFFER = "match offer"
HURRY_ACTION = "hurry action"
CHAT_ACTION = "chat action"

ACTIONS = (
    RESIGNATION,
    FLAG_CALL,
    DRAW_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    PAUSE_OFFER,
    RESUME_OFFER,
    SWITCH_OFFER,
    TAKEBACK_OFFER,
    MATCH_OFFER,
    HURRY_ACTION,
    CHAT_ACTION,
)
OFFERS = (
    DRAW_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    PAUSE_OFFER,
    RESUME_OFFER,
    SWITCH_OFFER,
    TAKEBACK_OFFER,
    MATCH_OFFER,
)
INGAME_ACTIONS = (
    RESIGNATION,
    FLAG_CALL,
    DRAW_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    PAUSE_OFFER,
    SWITCH_OFFER,
    HURRY_ACTION,
)

# A few nice to have boards
FEN_EMPTY = "4k3/8/8/8/8/8/8/4K3 w - - 0 1"
FEN_START = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"
FEN_START_960 = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w HAha - 0 1"

# Search values

hashfALPHA, hashfBETA, hashfEXACT, hashfBAD = range(4)

# Engine modes
NORMAL, ANALYZING, INVERSE_ANALYZING = range(3)

# Piece types

# BPAWN is a pawn that moves in the opposite direction
(
    EMPTY,
    PAWN,
    KNIGHT,
    BISHOP,
    ROOK,
    QUEEN,
    KING,
    HAWK,
    ELEPHANT,
    BPAWN,
    ASEAN_WBISHOP,
    ASEAN_BBISHOP,
    ASEAN_QUEEN,
) = range(13)

# Is sliding piece
sliders = [
    False,
    False,
    False,
    True,
    True,
    True,
    False,
    True,
    True,
    False,
    False,
    False,
    False,
]

# Piece signs
reprSign = ["", "P", "N", "B", "R", "Q", "K", "H", "E"]
reprSignMakruk = ["", "P", "N", "S", "R", "M", "K"]
reprSignSittuyin = ["", "P", "N", "S", "R", "F", "K"]
chr2Sign = {
    "k": KING,
    "q": QUEEN,
    "r": ROOK,
    "b": BISHOP,
    "n": KNIGHT,
    "p": PAWN,
    "m": QUEEN,
    "s": BISHOP,
    "f": QUEEN,
    "h": HAWK,
    "e": ELEPHANT,
}
chrU2Sign = {
    "K": KING,
    "Q": QUEEN,
    "R": ROOK,
    "B": BISHOP,
    "N": KNIGHT,
    "P": PAWN,
    "M": QUEEN,
    "S": BISHOP,
    "F": QUEEN,
    "H": HAWK,
    "E": ELEPHANT,
}

# Move values
(
    NORMAL_MOVE,
    QUEEN_CASTLE,
    KING_CASTLE,
    ENPASSANT,
    KNIGHT_PROMOTION,
    BISHOP_PROMOTION,
    ROOK_PROMOTION,
    QUEEN_PROMOTION,
    KING_PROMOTION,
    HAWK_PROMOTION,
    ELEPHANT_PROMOTION,
    HAWK_GATE,
    ELEPHANT_GATE,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
    DROP,
) = range(16)

PROMOTIONS = (
    KING_PROMOTION,
    QUEEN_PROMOTION,
    ROOK_PROMOTION,
    BISHOP_PROMOTION,
    KNIGHT_PROMOTION,
    HAWK_PROMOTION,
    ELEPHANT_PROMOTION,
)

GATINGS = (HAWK_GATE, ELEPHANT_GATE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)

# Algebraic notation types: Short, Long, Figure and Simpe
SAN, LAN, FAN, AN = range(4)
# Castling notation types: e.g., O-O, e1g1, e1h1
CASTLE_SAN, CASTLE_KK, CASTLE_KR = range(3)

FAN_PIECES = [
    ["", "♙", "♘", "♗", "♖", "♕", "♔", "H", "E", ""],
    ["", "♟", "♞", "♝", "♜", "♛", "♚", "h", "e", ""],
]

# Castling values
W_OO, W_OOO, B_OO, B_OOO = (2**i for i in range(4))
CAS_FLAGS = ((W_OOO, W_OO), (B_OOO, B_OO))
W_CASTLED, B_CASTLED = (2**i for i in range(2))

NB_OF_CASES = 64

# Cords types
(
    A1,
    B1,
    C1,
    D1,
    E1,
    F1,
    G1,
    H1,
    A2,
    B2,
    C2,
    D2,
    E2,
    F2,
    G2,
    H2,
    A3,
    B3,
    C3,
    D3,
    E3,
    F3,
    G3,
    H3,
    A4,
    B4,
    C4,
    D4,
    E4,
    F4,
    G4,
    H4,
    A5,
    B5,
    C5,
    D5,
    E5,
    F5,
    G5,
    H5,
    A6,
    B6,
    C6,
    D6,
    E6,
    F6,
    G6,
    H6,
    A7,
    B7,
    C7,
    D7,
    E7,
    F7,
    G7,
    H7,
    A8,
    B8,
    C8,
    D8,
    E8,
    F8,
    G8,
    H8,
) = range(64)

reprCord = [
    "a1",
    "b1",
    "c1",
    "d1",
    "e1",
    "f1",
    "g1",
    "h1",
    "a2",
    "b2",
    "c2",
    "d2",
    "e2",
    "f2",
    "g2",
    "h2",
    "a3",
    "b3",
    "c3",
    "d3",
    "e3",
    "f3",
    "g3",
    "h3",
    "a4",
    "b4",
    "c4",
    "d4",
    "e4",
    "f4",
    "g4",
    "h4",
    "a5",
    "b5",
    "c5",
    "d5",
    "e5",
    "f5",
    "g5",
    "h5",
    "a6",
    "b6",
    "c6",
    "d6",
    "e6",
    "f6",
    "g6",
    "h6",
    "a7",
    "b7",
    "c7",
    "d7",
    "e7",
    "f7",
    "g7",
    "h7",
    "a8",
    "b8",
    "c8",
    "d8",
    "e8",
    "f8",
    "g8",
    "h8",
]

reprFile = ["a", "b", "c", "d", "e", "f", "g", "h"]
reprRank = ["1", "2", "3", "4", "5", "6", "7", "8"]

# dictionary for giving the cord linked to the flattened index of a coordinate
cordDic = {}
for cord, name in enumerate(reprCord):
    cordDic[name] = cord

# User interface

# pagination in database game list
FIRST_PAGE = 0
PREV_PAGE = -1
NEXT_PAGE = 1

# Hint modes
OPENING, ENDGAME, HINT, SPY = ["opening", "endgame", "hint", "spy"]

# Sound settings
SOUND_MUTE, SOUND_BEEP, SOUND_SELECT, SOUND_URI = range(4)
COUNT_OF_SOUNDS = 13

# Brush types. Send piece object for Piece brush
CLEAR, ENPAS = range(2)

# Main menu items
GAME_MENU_ITEMS = (
    "save_game1",
    "save_game_as1",
    "share_game",
    "export_position1",
    "analyze_game1",
    "properties1",
    "close1",
)
ACTION_MENU_ITEMS = (
    "abort",
    "adjourn",
    "draw",
    "pause1",
    "resume1",
    "undo1",
    "call_flag",
    "resign",
    "ask_to_move",
)
VIEW_MENU_ITEMS = ("rotate_board1", "show_sidepanels", "hint_mode", "spy_mode")
EDIT_MENU_ITEMS = (
    "copy_pgn",
    "copy_fen",
)
MENU_ITEMS = GAME_MENU_ITEMS + ACTION_MENU_ITEMS + VIEW_MENU_ITEMS + EDIT_MENU_ITEMS

# Column name
COLUMN_ROW_RESET = "column_row_reset"

# Gtk icon name
GTK_ICON_VIEW_REFRESH = "view-refresh"

# Learn categories
LECTURE, LESSON, PUZZLE, ENDGAME = range(4)



import time
import logging

from gi.repository import Gtk, Gdk, Pango, GLib

from pychess.System import uistuff
from pychess.System.LogEmitter import logemitter


class InformationWindow:
    @classmethod
    def _init(cls):
        cls.tagToIter = {}
        cls.tagToPage = {}
        cls.pathToPage = {}
        cls.tagToTime = {}

        cls.window = Gtk.Window()
        cls.window.set_title(_("PyChess Information Window"))
        cls.window.set_border_width(12)
        cls.window.set_icon_name("pychess")
        uistuff.keepWindowSize("logdialog", cls.window, (640, 480))
        mainHBox = Gtk.HBox()
        mainHBox.set_spacing(6)
        cls.window.add(mainHBox)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.IN)
        mainHBox.pack_start(sw, False, True, 0)
        cls.treeview = Gtk.TreeView(Gtk.TreeStore(str))
        cls.treeview.append_column(
            Gtk.TreeViewColumn("", Gtk.CellRendererText(), text=0)
        )
        cls.treeview.set_headers_visible(False)
        cls.treeview.get_selection().set_mode(Gtk.SelectionMode.BROWSE)
        sw.add(cls.treeview)
        cls.pages = Gtk.Notebook()
        cls.pages.set_show_tabs(False)
        cls.pages.set_show_border(False)
        mainHBox.pack_start(cls.pages, True, True, 0)

        mainHBox.show_all()

        def selectionChanged(selection):
            treestore, iter = selection.get_selected()
            if iter:
                child = cls.pathToPage[treestore.get_path(iter).to_string()]["child"]
                cls.pages.set_current_page(cls.pages.page_num(child))

        cls.treeview.get_selection().connect("changed", selectionChanged)

    @classmethod
    def show(cls):
        cls.window.show()

    @classmethod
    def hide(cls):
        cls.window.hide()

    @classmethod
    def newMessage(cls, tag, timestamp, message, importance):
        def _newMessage(cls, tag, timestamp, message, importance):
            textview = cls._getPageFromTag(tag)["textview"]

            if tag not in cls.tagToTime or timestamp - cls.tagToTime[tag] >= 1:
                t = time.strftime("%H:%M:%S", time.localtime(timestamp))
                textview.get_buffer().insert_with_tags_by_name(
                    textview.get_buffer().get_end_iter(),
                    "\n{}\n{}\n".format(t, "-" * 60),
                    str(logging.INFO),
                )
                cls.tagToTime[tag] = timestamp

            if not message.endswith("\n"):
                message = "%s\n" % message
            textview.get_buffer().insert_with_tags_by_name(
                textview.get_buffer().get_end_iter(), message, str(importance)
            )

        GLib.idle_add(_newMessage, cls, tag, timestamp, message, importance)

    @classmethod
    def _createPage(cls, parent_iter, tag):
        name = tag[-1]
        if isinstance(name, int):
            name = str(name)
        iter = cls.treeview.get_model().append(parent_iter, (name,))
        cls.tagToIter[tag] = iter

        widgets = uistuff.GladeWidgets("findbar.glade")
        frame = widgets["frame"]
        frame.unparent()
        frame.show_all()

        uistuff.keepDown(widgets["scrolledwindow"])
        textview = widgets["textview"]
        tb = textview.get_buffer()
        tb.create_tag(str(logging.DEBUG), family="Monospace")
        tb.create_tag(str(logging.INFO), family="Monospace", weight=Pango.Weight.BOLD)
        tb.create_tag(str(logging.WARNING), family="Monospace", foreground="red")
        tb.create_tag(
            str(logging.ERROR),
            family="Monospace",
            weight=Pango.Weight.BOLD,
            foreground="red",
        )
        tb.create_tag(
            str(logging.CRITICAL),
            family="Monospace",
            weight=Pango.Weight.BOLD,
            foreground="red",
        )

        findbar = widgets["findbar"]
        findbar.hide()
        # Make searchEntry and "out of label" share height with the buttons
        widgets["prevButton"].connect(
            "size-allocate",
            lambda w, alloc: widgets["searchEntry"].set_size_request(-1, alloc.height)
            or widgets["outofLabel"].set_size_request(-1, alloc.height - 2),
        )

        # Make "out of label" more visually distinct
        uistuff.makeYellow(widgets["outofLabel"])
        widgets["outofLabel"].hide()

        widgets["closeButton"].connect("clicked", lambda w: widgets["findbar"].hide())

        # Connect showing/hiding of the findbar
        cls.window.connect("key-press-event", cls.onTextviewKeypress, widgets)
        widgets["findbar"].connect("key-press-event", cls.onFindbarKeypress)

        widgets["searchEntry"].connect("changed", cls.onSearchChanged, widgets)

        widgets["prevButton"].connect("clicked", lambda w: cls.searchJump(-1, widgets))
        widgets["nextButton"].connect("clicked", lambda w: cls.searchJump(1, widgets))

        cls.pages.append_page(frame, None)
        page = {"child": frame, "textview": textview}
        cls.tagToPage[tag] = page
        cls.pathToPage[cls.treeview.get_model().get_path(iter).to_string()] = page

        cls.treeview.expand_all()

    @classmethod
    def _getPageFromTag(cls, tag):
        if isinstance(tag, list):
            tag = tuple(tag)
        elif not isinstance(tag, tuple):
            tag = (tag,)

        if tag in cls.tagToPage:
            return cls.tagToPage[tag]

        for i in range(len(tag) - 1):
            subtag = tag[: -i - 1]
            if subtag in cls.tagToIter:
                newtag = subtag + (tag[len(subtag)],)
                iter = cls.tagToIter[subtag]
                cls._createPage(iter, newtag)
                return cls._getPageFromTag(tag)

        cls._createPage(None, tag[:1])
        return cls._getPageFromTag(tag)

    @classmethod
    def onSearchChanged(cls, search_entry, widgets):
        pattern = search_entry.get_text().lower()
        widgets["outofLabel"].props.visible = bool(pattern)
        if not pattern:
            return

        text = widgets["textview"].get_buffer().props.text.lower()

        widgets["outofLabel"].hits = []
        widgets["outofLabel"].searchCurrent = -1
        i = -len(pattern)
        while True:
            i = text.find(pattern, i + len(pattern))
            if i != -1:
                widgets["outofLabel"].hits.append(i)
            else:
                break

        cls.searchJump(1, widgets)

    @classmethod
    def searchJump(cls, count, widgets):
        if not hasattr(widgets["outofLabel"], "hits"):
            return
        amount = len(widgets["outofLabel"].hits)
        if not amount:
            widgets["outofLabel"].set_text("0 %s 0" % _("of"))
        else:
            widgets["outofLabel"].searchCurrent += count
            current = widgets["outofLabel"].searchCurrent % amount
            widgets["outofLabel"].set_text("%d %s %d" % (current + 1, _("of"), amount))
            goto = widgets["outofLabel"].hits[current]
            iter0 = widgets["textview"].get_buffer().get_iter_at_offset(goto)
            length = len(widgets["searchEntry"].get_text())
            iter1 = widgets["textview"].get_buffer().get_iter_at_offset(goto + length)
            widgets["textview"].get_buffer().select_range(iter0, iter1)
            widgets["textview"].scroll_to_iter(iter0, 0.2, False, 0.5, 0.5)

    @classmethod
    def onTextviewKeypress(cls, textview, event, widgets):
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            if event.keyval in (ord("f"), ord("F")):
                widgets["findbar"].props.visible = not widgets["findbar"].props.visible
                if widgets["findbar"].props.visible:
                    signal = widgets["searchEntry"].connect_after(
                        "draw",
                        lambda w, e: w.grab_focus()
                        or widgets["searchEntry"].disconnect(signal),
                    )

    @classmethod
    def onFindbarKeypress(cls, findbar, event):
        if Gdk.keyval_name(event.keyval) == "Escape":
            findbar.props.visible = False

    ################################################################################
    # Add early messages and connect for new                                       #
    ################################################################################


InformationWindow._init()


def addMessage(emitter, message):
    task, timestamp, message, type = message
    InformationWindow.newMessage(task, timestamp, message, type)


for message in logemitter.messages:
    addMessage(logemitter, message)
logemitter.messages = None

logemitter.connect("logged", addMessage)

################################################################################
# External functions                                                           #
################################################################################

destroy_funcs = []


def add_destroy_notify(func):
    destroy_funcs.append(func)


def _destroy_notify(widget, *args):
    [func() for func in destroy_funcs]
    return True


InformationWindow.window.connect("delete-event", _destroy_notify)


def show():
    InformationWindow.show()


def hide():
    InformationWindow.hide()


if __name__ == "__main__":
    show()
    InformationWindow.window.connect("delete-event", Gtk.main_quit)
    Gtk.main()



from collections import defaultdict

from pychess.Players.Player import (
    Player,
    PlayerIsDead,
    PassInterrupt,
    TurnInterrupt,
    GameEnded,
)
from pychess.Utils.Move import parseSAN, toAN
from pychess.Utils.lutils.lmove import ParsingError
from pychess.Utils.Offer import Offer
from pychess.Utils.const import (
    REMOTE,
    UNFINISHED_STATES,
    CHAT_ACTION,
    CASTLE_KK,
    FISCHERRANDOMCHESS,
    CASTLE_SAN,
    TAKEBACK_OFFER,
)
from pychess.System.Log import log


class ICPlayer(Player):
    __type__ = REMOTE

    def __init__(self, gamemodel, ichandle, gameno, color, name, icrating=None):
        Player.__init__(self)
        self.offers = {}
        self.setName(name)
        self.ichandle = ichandle
        self.icrating = icrating
        self.color = color
        self.gameno = gameno
        self.gamemodel = gamemodel
        self.pass_interrupt = False
        self.turn_interrupt = False

        self.connection = connection = self.gamemodel.connection

        self.connections = connections = defaultdict(list)
        connections[connection.om].append(
            connection.om.connect("onOfferAdd", self.__onOfferAdd)
        )
        connections[connection.om].append(
            connection.om.connect("onOfferRemove", self.__onOfferRemove)
        )
        connections[connection.om].append(
            connection.om.connect("onOfferDeclined", self.__onOfferDeclined)
        )
        connections[connection.cm].append(
            connection.cm.connect("privateMessage", self.__onPrivateMessage)
        )

        self.cid = self.gamemodel.connect_after(
            "game_terminated", self.on_game_terminated
        )

    def on_game_terminated(self, model):
        self.gamemodel.disconnect(self.cid)

    def getICHandle(self):
        return self.name

    @property
    def time(self):
        return self.gamemodel.timemodel.getPlayerTime(self.color)

    @property
    def move_queue(self):
        return self.gamemodel.ficsgame.move_queue

    # Handle signals from the connection

    def __onOfferAdd(self, om, offer):
        if (
            self.gamemodel.status in UNFINISHED_STATES
            and not self.gamemodel.isObservationGame()
        ):
            log.debug(
                "ICPlayer.__onOfferAdd: emitting offer: self.gameno={} self.name={} {}".format(
                    self.gameno, self.name, offer
                )
            )
            self.offers[offer.index] = offer
            self.emit("offer", offer)

    def __onOfferDeclined(self, om, offer):
        for offer_ in list(self.gamemodel.offers.keys()):
            if offer.type == offer_.type:
                offer.param = offer_.param
        log.debug("ICPlayer.__onOfferDeclined: emitting decline for %s" % offer)
        self.emit("decline", offer)

    def __onOfferRemove(self, om, offer):
        if offer.index in self.offers:
            log.debug(
                "ICPlayer.__onOfferRemove: emitting withdraw: \
                      self.gameno={} self.name={} {}".format(
                    self.gameno, self.name, offer
                )
            )
            # self.emit("withdraw", self.offers[offer.index])
            del self.offers[offer.index]

    def __onPrivateMessage(self, cm, name, title, isadmin, text):
        if name == self.ichandle:
            self.emit("offer", Offer(CHAT_ACTION, param=text))

    def __disconnect(self):
        log.debug("ICPlayer.__disconnect: %s" % self.name)
        if self.connections is None:
            return
        for obj in self.connections:
            for handler_id in self.connections[obj]:
                if obj.handler_is_connected(handler_id):
                    obj.disconnect(handler_id)
        self.connections = None

    def end(self, status=None, reason=None):
        log.debug("ICPlayer.end: %s" % self.name)
        self.__disconnect()
        self.move_queue.put_nowait("end")

    def kill(self, reason):
        self.__disconnect()
        self.move_queue.put_nowait("del")

    # Send the player move updates

    async def makeMove(self, board1, move, board2):
        log.debug(
            "ICPlayer.makemove: id(self)=%d self=%s move=%s board1=%s board2=%s"
            % (id(self), self, move, board1, board2)
        )
        if board2 and not self.gamemodel.isObservationGame():
            # TODO: Will this work if we just always use CASTLE_SAN?
            castle_notation = CASTLE_KK
            if board2.variant == FISCHERRANDOMCHESS:
                castle_notation = CASTLE_SAN
            self.connection.bm.sendMove(
                toAN(board2, move, castleNotation=castle_notation)
            )
            # wait for fics to send back our move we made
            item = await self.move_queue.get()
            log.debug("ICPlayer.makeMove: fics sent back the move we made")

        item = await self.move_queue.get()
        try:
            if item == "end":
                log.debug("ICPlayer.makeMove got: end")
                raise GameEnded
            elif item == "del":
                log.debug("ICPlayer.makeMove got: del")
                raise PlayerIsDead
            elif item == "stm":
                log.debug("ICPlayer.makeMove got: stm")
                self.turn_interrupt = False
                raise TurnInterrupt
            elif item == "fen":
                log.debug("ICPlayer.makeMove got: fen")
                self.turn_interrupt = False
                raise TurnInterrupt
            elif item == "pass":
                log.debug("ICPlayer.makeMove got: pass")
                self.pass_interrupt = False
                raise PassInterrupt

            gameno, ply, curcol, lastmove, fen, wname, bname, wms, bms = item
            log.debug(
                "ICPlayer.makeMove got: {} {} {} {}".format(
                    gameno, ply, curcol, lastmove
                )
            )
            self.gamemodel.update_board(
                gameno, ply, curcol, lastmove, fen, wname, bname, wms, bms
            )

            if self.turn_interrupt:
                self.turn_interrupt = False
                raise TurnInterrupt

            if self.pass_interrupt:
                self.pass_interrupt = False
                raise PassInterrupt

            if ply < board1.ply:
                # This should only happen in an observed game
                board1 = self.gamemodel.getBoardAtPly(max(ply - 1, 0))
            log.debug(
                "ICPlayer.makemove: id(self)=%d self=%s from queue got: ply=%d sanmove=%s"
                % (id(self), self, ply, lastmove)
            )

            try:
                move = parseSAN(board1, lastmove)
                log.debug(
                    "ICPlayer.makemove: id(self)=%d self=%s parsed move=%s"
                    % (id(self), self, move)
                )
            except ParsingError:
                raise
            return move
        finally:
            log.debug(
                "ICPlayer.makemove: id(self)=%d self=%s returning move=%s"
                % (id(self), self, move)
            )

    # Interacting with the player

    def pause(self):
        pass

    def resume(self):
        pass

    def setBoard(self, fen):
        # setBoard will currently only be called for ServerPlayer when starting
        # to observe some game. In this case FICS already knows how the board
        # should look, and we don't need to set anything
        pass

    def playerUndoMoves(self, movecount, gamemodel):
        log.debug(
            "ICPlayer.playerUndoMoves: id(self)=%d self=%s, undoing movecount=%d"
            % (id(self), self, movecount)
        )
        # If current player has changed so that it is no longer us to move,
        # We raise TurnInterrupt in order to let GameModel continue the game
        if movecount % 2 == 1 and gamemodel.curplayer != self:
            log.debug(
                "ICPlayer.playerUndoMoves: set self.turn_interrupt = True %s"
                % self.name
            )
            if self.connection.ICC:
                self.move_queue.put_nowait("stm")
            else:
                self.turn_interrupt = True
        if movecount % 2 == 0 and gamemodel.curplayer == self:
            log.debug(
                "ICPlayer.playerUndoMoves: set self.pass_interrupt = True %s"
                % self.name
            )
            if self.connection.ICC:
                self.move_queue.put_nowait("pass")
            else:
                self.pass_interrupt = True

    def resetPosition(self):
        """Used in observed examined games f.e. when LectureBot starts another example"""
        self.turn_interrupt = True

    def putMessage(self, text):
        self.connection.cm.tellPlayer(self.ichandle, text)

    # Offer handling

    def offerRematch(self):
        if self.gamemodel.timed:
            minimum = int(self.gamemodel.timemodel.intervals[0][0]) / 60
            inc = self.gamemodel.timemodel.gain
        else:
            minimum = 0
            inc = 0
        self.connection.om.challenge(
            self.ichandle,
            self.gamemodel.ficsgame.game_type,
            minimum,
            inc,
            self.gamemodel.ficsgame.rated,
        )

    def offer(self, offer):
        log.debug(f"ICPlayer.offer: self={repr(self)} {offer}")
        if offer.type == TAKEBACK_OFFER:
            # only 1 outstanding takeback offer allowed on FICS, so remove any of ours
            for index in list(self.offers.keys()):
                if self.offers[index].type == TAKEBACK_OFFER:
                    log.debug(f"ICPlayer.offer: del self.offers[{index}] {offer}")
                    del self.offers[index]
        self.connection.om.offer(offer)

    def offerDeclined(self, offer):
        log.debug("ICPlayer.offerDeclined: sending decline for %s" % offer)
        self.connection.om.decline(offer)

    def offerWithdrawn(self, offer):
        pass

    def offerError(self, offer, error):
        pass

    def observe(self):
        self.connection.client.run_command("observe %s" % self.ichandle)



"""Light Brigade Chess Variant"""

from pychess.Utils.const import (
    QUEEN_PROMOTION,
    KNIGHT_PROMOTION,
    LIGHTBRIGADECHESS,
    VARIANTS_OTHER_NONSTANDARD,
)
from pychess.Utils.Board import Board

LIGHTBRIGADESTART = "nnnnknnn/pppppppp/8/8/8/8/PPPPPPPP/1Q1QK1Q1 w - - 0 1"


class LightbrigadeBoard(Board):
    variant = LIGHTBRIGADECHESS
    __desc__ = _(
        "Variant explained at https://www.chessvariants.com/rules/charge-of-the-light-brigade"
    )
    name = _("Charge of the light brigade")
    cecp_name = "light-brigade"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    PROMOTIONS = (QUEEN_PROMOTION, KNIGHT_PROMOTION)

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=LIGHTBRIGADESTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



# AsymmetricRandom Chess

import random

from pychess.Utils.const import VARIANTS_SHUFFLE, ASYMMETRICRANDOMCHESS
from pychess.Utils.Board import Board


class AsymmetricRandomBoard(Board):
    variant = ASYMMETRICRANDOMCHESS
    __desc__ = _(
        "FICS wild/4: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "* Randomly chosen pieces (two queens or three rooks possible)\n"
        + "* Exactly one king of each color\n"
        + "* Pieces placed randomly behind the pawns, SUBJECT TO THE CONSTRAINT THAT THE BISHOPS ARE BALANCED\n"
        + "* No castling\n"
        + "* Black's arrangement DOES NOT mirrors white's"
    )
    name = _("Asymmetric Random")
    cecp_name = "wild/4"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.asymmetricrandom_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def asymmetricrandom_start(self):
        white = random.sample(("r", "n", "b", "q") * 16, 7)
        white.append("k")
        black = white[:]
        random.shuffle(white)
        random.shuffle(black)

        # balance the bishops (put them on equal numbers of dark and light squares)
        whitedarkbishops = 0
        whitelightbishops = 0
        for index, piece in enumerate(white):
            if piece == "b":
                if index % 2 == 0:  # even numbered square on the A rank are dark
                    whitedarkbishops += 1
                else:
                    whitelightbishops += 1
        blackdarkbishops = 0
        blacklightbishops = 0
        blackbishoprandomindexstack = []
        for index, piece in enumerate(black):
            if piece == "b":
                if index % 2 == 1:  # odd numbered squares on the H rank are dark
                    blackdarkbishops += 1
                else:
                    blacklightbishops += 1
                blackbishoprandomindexstack.append(index)
        random.shuffle(blackbishoprandomindexstack)

        class RandomEnumeratePieces:
            def __init__(self, pieces):
                self.pieces = pieces[:]
                self.randomindexstack = list(range(8))
                random.shuffle(self.randomindexstack)

            def __iter__(self):
                return self

            def __next__(self):
                if not self.randomindexstack:
                    raise StopIteration
                else:
                    randomindex = self.randomindexstack.pop()
                    return randomindex, self.pieces[randomindex]

        while (whitedarkbishops != blackdarkbishops) or (
            whitelightbishops != blacklightbishops
        ):
            bishopindex = blackbishoprandomindexstack.pop()
            for index, piece in RandomEnumeratePieces(black):
                if piece != "b":
                    if (
                        (blackdarkbishops > whitedarkbishops)
                        and (bishopindex % 2 == 1)
                        and (index % 2 == 0)
                    ):
                        black[bishopindex] = piece
                        black[index] = "b"
                        blacklightbishops += 1
                        blackdarkbishops = (
                            blackdarkbishops > 0 and (blackdarkbishops - 1) or 0
                        )
                        break
                    elif (
                        (blacklightbishops > whitelightbishops)
                        and (bishopindex % 2 == 0)
                        and (index % 2 == 1)
                    ):
                        black[bishopindex] = piece
                        black[index] = "b"
                        blackdarkbishops += 1
                        blacklightbishops = (
                            blacklightbishops > 0 and (blacklightbishops - 1) or 0
                        )
                        break

        tmp = (
            "".join(black)
            + "/pppppppp/8/8/8/8/PPPPPPPP/"
            + "".join(white).upper()
            + " w - - 0 1"
        )

        return tmp


if __name__ == "__main__":
    Board = AsymmetricRandomBoard(True)
    for i in range(10):
        print(Board.asymmetricrandom_start())



import os
import sys

try:
    import gi
except ImportError:
    print("ERROR: gst_player requires pygobject to be installed.")
    sys.exit(1)

from gi.repository import GLib

try:
    gi.require_version("Gst", "1.0")
    from gi.repository import Gst
except (ImportError, ValueError) as err:
    print("ERROR: Unable to import gstreamer.\n%s" % err)
    sys.exit(1)

if not Gst.init_check(None):
    print("ERROR: Unable to initialize gstreamer.")
    sys.exit(1)

player = Gst.ElementFactory.make("playbin", "player")
if player is None:
    print('ERROR: Gst.ElementFactory.make("playbin", "player") failed')
    sys.exit(1)


def on_message(bus, message):
    if message.type == Gst.MessageType.ERROR:
        player.set_state(Gst.State.NULL)
        simpleMessage, advMessage = message.parse_error()
        print(f"Gstreamer error '{simpleMessage}': {advMessage}")
    elif message.type == Gst.MessageType.EOS:
        player.set_state(Gst.State.NULL)
    return True


fakesink = Gst.ElementFactory.make("fakesink", "fakesink")
player.set_property("video-sink", fakesink)
bus = player.get_bus()
bus.add_signal_watch()
bus.connect("message", on_message)


def play(loop):
    line = sys.stdin.readline().strip()
    if not line:
        loop.quit()
        return

    if os.path.isfile(line):
        player.set_state(Gst.State.NULL)
        player.set_property("uri", "file://" + line)
        player.set_state(Gst.State.PLAYING)
    else:
        print("file not found:", line)
        player.set_state(Gst.State.NULL)
    return True


loop = GLib.MainLoop()
GLib.idle_add(play, loop)
loop.run()



import asyncio
from io import StringIO

from gi.repository import Gtk

from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import WHITE, BLACK, LOCAL, RUNNING, LECTURE
from pychess.Utils.LearnModel import LearnModel
from pychess.Utils.TimeModel import TimeModel
from pychess.Utils.Move import parseAny
from pychess.Players.Human import Human
from pychess.perspectives import perspective_manager
from pychess.Savers import fen as fen_loader
from pychess.System import conf

__title__ = _("Lectures")

__icon__ = addDataPrefix("glade/panel_book.svg")

__desc__ = _("Study FICS lectures offline")


LECTURES = (
    ("lec1.txt", "2...Qh4+ against the King's Gambit", "toddmf"),
    ("lec2.txt", "Tactics Training lesson 1# 'Back rank weakness'", "knackie"),
    ("lec3.txt", "Denker's Favorite Game", "toddmf"),
    ("lec4.txt", "Introduction to the 2.Nc3 Caro-Kann", "KayhiKing"),
    ("lec5.txt", "Tactics Training lesson 2# 'Discovered Attack'", "knackie"),
    ("lec6.txt", "King's Indian Attack vs. the French", "cissmjg"),
    ("lec7.txt", "Rook vs Pawn endgames", "toddmf"),
    ("lec8.txt", "The Stonewall Attack", "MBDil"),
    ("lec9.txt", "Tactics Training lesson 3# 'Enclosed Kings'", "knackie"),
    ("lec10.txt", "The Steinitz Variation of the French Defense", "Seipman"),
    ("lec11.txt", "A draw against a Grandmaster", "talpa"),
    ("lec12.txt", "Tactics Training lesson 4# 'King in the centre'", "knackie"),
    ("lec13.txt", "The Modern Defense", "GMDavies"),
    (
        "lec14.txt",
        "Tactics Training lesson 5# 'Pulling the king to the open'",
        "knackie",
    ),
    ("lec15.txt", "King's Indian Attack vs. the Caro-Kann", "cissmjg"),
    ("lec16.txt", "Introduction to Bughouse", "Tecumseh"),
    ("lec17.txt", "Refuting the Milner-Barry Gambit in the French Defense", "Kabal"),
    ("lec18.txt", "Tactics Training lesson 6# 'Mating Attack'", "knackie"),
    ("lec19.txt", "Closed Sicilian Survey, part 1", "Oren"),
    ("lec20.txt", "Hypermodern Magic - A study of the central blockade", "Bahamut"),
    ("lec21.txt", "Tactics Training lesson 7# 'Opening / Closing Files'", "knackie"),
    ("lec22.txt", "Thoughts on the Refutation of the Milner-Barry", "knackie"),
    (
        "lec23.txt",
        "Tactics Training lesson 8# 'Opening / Closing Diagonals'",
        "knackie",
    ),
    ("lec24.txt", "King's Indian Attack vs. Other Defenses", "cissmjg"),
    ("lec25.txt", "Basic Pawn Endings I", "DAV"),
    ("lec26.txt", "Giuoco Piano", "afw"),
    ("lec27.txt", "Tactics Training lesson 9# 'Long Diagonals'", "knackie"),
    ("lec28.txt", "Secrets of the Flank Attack", "Shidinov"),
    ("lec29.txt", "Mating Combinations", "Kabal"),
    ("lec30.txt", "Basic Pawn Endings II", "DAV"),
    ("lec31.txt", "Grandmaster Knezevic's first FICS lecture", "toddmf"),
)


class Sidepanel:
    def load(self, persp):
        self.persp = persp
        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.tv = Gtk.TreeView()

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Title"), renderer, text=1)
        self.tv.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Author"), renderer, text=2)
        self.tv.append_column(column)

        self.tv.connect("row-activated", self.row_activated)

        self.store = Gtk.ListStore(str, str, str)

        for file_name, title, author in LECTURES:
            self.store.append([file_name, title, author])

        self.tv.set_model(self.store)
        self.tv.get_selection().set_mode(Gtk.SelectionMode.BROWSE)
        self.tv.set_cursor(conf.get("learncombo%s" % LECTURE))

        scrollwin = Gtk.ScrolledWindow()
        scrollwin.add(self.tv)
        scrollwin.show_all()

        self.box.pack_start(scrollwin, True, True, 0)
        self.box.show_all()

        return self.box

    def row_activated(self, widget, path, col):
        if path is None:
            return
        else:
            filename = LECTURES[path[0]][0]
            conf.set("categorycombo", LECTURE)
            from pychess.widgets.TaskerManager import learn_tasker

            learn_tasker.learn_combo.set_active(path[0])
            start_lecture_from(filename)


def start_lecture_from(filename, index=None):
    if index is None:
        index = 0

    # connection.client.run_command("examine")
    timemodel = TimeModel(0, 0)
    gamemodel = LearnModel(timemodel)
    gamemodel.set_learn_data(LECTURE, filename, index)

    white_name = black_name = "PyChess"
    p0 = (LOCAL, Human, (WHITE, white_name), white_name)
    p1 = (LOCAL, Human, (BLACK, black_name), black_name)

    def on_game_started(gamemodel):
        perspective.activate_panel("chatPanel")

    gamemodel.connect("game_started", on_game_started)

    perspective = perspective_manager.get_perspective("games")
    asyncio.create_task(perspective.generalStart(gamemodel, p0, p1))

    def lecture_steps(lecture_file):
        with open(lecture_file) as f:
            yield from f
        return

    lecture_file = addDataPrefix("learn/lectures/%s" % filename)
    steps = lecture_steps(lecture_file)

    async def coro(gamemodel, steps):
        exit_lecture = False
        inside_bsetup = False
        paused = False
        moves_played = 0

        (
            KIBITZ,
            BACKWARD,
            BSETUP,
            BSETUP_DONE,
            FEN,
            TOMOVE,
            WCASTLE,
            BCASTLE,
            WNAME,
            BNAME,
            REVERT,
            WAIT,
            MOVE,
        ) = range(13)

        while True:
            try:
                step = next(steps)
                print(step)

                parts = step.strip().split()

                command = None
                param = ""

                if parts[0].lower() in ("k", "ki", "kib", "kibitz"):
                    command = KIBITZ
                    param = " ".join(parts[1:])
                elif parts[0] == "back":
                    command = BACKWARD
                    param = int(parts[1]) if len(parts) > 1 else 1
                elif parts[0] == "bsetup":
                    if len(parts) == 1:
                        command = BSETUP
                    else:
                        if parts[1] == "done":
                            command = BSETUP_DONE
                        elif parts[1] == "fen":
                            command = FEN
                            param = parts[2]
                        elif parts[1] == "tomove":
                            command = TOMOVE
                            param = "w" if parts[2].lower()[0] == "w" else "b"
                        elif parts[1] == "wcastle":
                            command = WCASTLE
                            param = parts[2]
                        elif parts[1] == "bcastle":
                            command = BCASTLE
                            param = parts[2]
                elif parts[0] == "tomove":
                    command = TOMOVE
                    param = "w" if parts[1].lower()[0] == "w" else "b"
                elif parts[0] == "wname":
                    command = WNAME
                    param = parts[1]
                elif parts[0] == "bname":
                    command = BNAME
                    param = parts[1]
                elif parts[0] == "revert":
                    command = REVERT
                elif len(parts) == 1 and parts[0].isdigit():
                    command = WAIT
                    param = int(parts[0])
                else:
                    command = MOVE
                    param = parts[0]

                if not inside_bsetup and command == BSETUP:
                    inside_bsetup = True
                    pieces = ""
                    color = ""
                    castl = ""
                    ep = ""
                elif inside_bsetup and command == BSETUP_DONE:
                    inside_bsetup = False

                wait_sec = int(param) if command == WAIT else 2

                if inside_bsetup:
                    wait_sec = -1

                while wait_sec >= 0:
                    if gamemodel.lecture_exit_event.is_set():
                        gamemodel.lecture_exit_event.clear()
                        exit_lecture = True
                        break

                    if gamemodel.lecture_skip_event.is_set():
                        gamemodel.lecture_skip_event.clear()
                        paused = False
                        break

                    if gamemodel.lecture_pause_event.is_set():
                        gamemodel.lecture_pause_event.clear()
                        paused = True

                    await asyncio.sleep(0.1)
                    if not paused:
                        wait_sec = wait_sec - 0.1

                if exit_lecture:
                    gamemodel.players[0].putMessage("Lecture exited.")
                    break

                if command != WAIT:
                    if command == KIBITZ:
                        gamemodel.players[0].putMessage(param)
                    if command == BACKWARD:
                        gamemodel.undoMoves(param)
                        moves_played -= param
                    if command == MOVE:
                        board = gamemodel.getBoardAtPly(gamemodel.ply)
                        move = parseAny(board, param)
                        gamemodel.curplayer.move_queue.put_nowait(move)
                        moves_played += 1
                    elif command == REVERT:
                        gamemodel.undoMoves(moves_played)
                        moves_played = 0
                    elif command == BNAME:
                        gamemodel.players[BLACK].name = param
                        gamemodel.emit("players_changed")
                    elif command == WNAME:
                        gamemodel.players[WHITE].name = param
                        gamemodel.emit("players_changed")
                    elif command == FEN:
                        pieces = param
                    elif command == TOMOVE:
                        color = param
                    elif command == WCASTLE:
                        if param == "both":
                            castl += "KQ"
                        elif param == "kside":
                            castl += "K"
                        elif param == "qside":
                            castl += "Q"
                    elif command == BCASTLE:
                        if param == "both":
                            castl += "kq"
                        elif param == "kside":
                            castl += "k"
                        elif param == "qside":
                            castl += "q"
                    elif command == BSETUP_DONE:
                        if not castl:
                            castl = "-"
                        if not ep:
                            ep = "-"
                        fen = f"{pieces} {color} {castl} {ep} 0 1"

                        curplayer = gamemodel.curplayer
                        gamemodel.status = RUNNING
                        gamemodel.loadAndStart(
                            StringIO(fen), fen_loader, 0, -1, first_time=False
                        )
                        curplayer.move_queue.put_nowait("int")
                        gamemodel.emit("game_started")
                        moves_played = 0

            except StopIteration:
                # connection.client.run_command("kibitz That concludes this lecture.")
                break

    asyncio.create_task(coro(gamemodel, steps))



import ast

from gi.repository import GLib, Gdk, Gtk, GObject

from pychess.Utils.const import chr2Sign, WHITE, BLACK, NORMALCHESS
from pychess.Utils.Piece import Piece
from pychess.Utils.SetupModel import SetupModel, SetupPlayer
from pychess.System import uistuff
from pychess.System.prefix import addDataPrefix
from pychess.widgets.BoardControl import BoardControl
from pychess.widgets.gameinfoDialog import on_pick_date
from pychess.widgets.PieceWidget import PieceWidget
from pychess.widgets import mainwindow
from pychess.Variants import name2variant

TAG_FILTER, MATERIAL_FILTER, PATTERN_FILTER, NONE, RULE, SEQUENCE, STREAK = range(7)


def formatted(q):
    """Simplified textual representation of query"""
    q = "%s" % q
    return q[1:-1].replace("'", "")


__title__ = _("Filters")

__icon__ = addDataPrefix("glade/panel_filter.svg")

__desc__ = _("Filters panel can filter game list by various conditions")


class FilterPanel(Gtk.TreeView):
    def __init__(self, persp):
        GObject.GObject.__init__(self)
        self.persp = persp
        self.filtered = False
        self.widgets = uistuff.GladeWidgets("PyChess.glade")

        # Build variant combo model
        variant_store = Gtk.ListStore(str, int)

        for name, variant in sorted(name2variant.items()):
            variant_store.append((name, variant.variant))

        self.widgets["variant"].set_model(variant_store)
        renderer_text = Gtk.CellRendererText()
        self.widgets["variant"].pack_start(renderer_text, True)
        self.widgets["variant"].add_attribute(renderer_text, "text", 0)

        # Connect date_from and date_to to corresponding calendars
        self.widgets["date_from_button"].connect(
            "clicked", on_pick_date, self.widgets["date_from"]
        )
        self.widgets["date_to_button"].connect(
            "clicked", on_pick_date, self.widgets["date_to"]
        )

        # Add piece widgets to dialog *_dock containers on material tab
        self.dialog = self.widgets["filter_dialog"]
        self.dialog.set_transient_for(mainwindow())

        def hide(widget, event):
            widget.hide()
            return True

        self.dialog.connect("delete-event", hide)

        for piece in "qrbnp":
            dock = "w%s_dock" % piece
            self.widgets[dock].add(PieceWidget(Piece(WHITE, chr2Sign[piece])))
            self.widgets[dock].get_child().show()

            dock = "b%s_dock" % piece
            self.widgets[dock].add(PieceWidget(Piece(BLACK, chr2Sign[piece])))
            self.widgets[dock].get_child().show()

            dock = "moved_%s_dock" % piece
            self.widgets[dock].add(PieceWidget(Piece(BLACK, chr2Sign[piece])))
            self.widgets[dock].get_child().show()

            dock = "captured_%s_dock" % piece
            self.widgets[dock].add(PieceWidget(Piece(BLACK, chr2Sign[piece])))
            self.widgets[dock].get_child().show()

        piece = "k"
        dock = "moved_%s_dock" % piece
        self.widgets[dock].add(PieceWidget(Piece(BLACK, chr2Sign[piece])))
        self.widgets[dock].get_child().show()

        self.widgets["copy_sub_fen"].connect("clicked", self.on_copy_sub_fen)
        self.widgets["paste_sub_fen"].connect("clicked", self.on_paste_sub_fen)

        # We will store our filtering queries in a ListStore
        # column 0: query as text
        # column 1: query dict
        # column 2: filter type (NONE, TAG_FILTER or MATERIAL_FILTER or PATTERN_FILTER)
        # column 3: row type (RULE, SEQUENCE, STREAK)
        self.treestore = Gtk.TreeStore(str, object, int, int)

        self.set_model(self.treestore)

        self.set_headers_visible(True)
        self.set_grid_lines(Gtk.TreeViewGridLines.HORIZONTAL)

        column = Gtk.TreeViewColumn(_("Filter"), Gtk.CellRendererText(), text=0)
        column.set_min_width(80)
        self.append_column(column)

        self.columns_autosize()

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        sw.add(self)

        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.box.pack_start(sw, True, True, 0)

        # Add buttons
        toolbar = Gtk.Toolbar()

        editButton = Gtk.ToolButton(stock_id=Gtk.STOCK_EDIT)
        editButton.set_tooltip_text(_("Edit selected filter"))
        editButton.connect("clicked", self.on_edit_clicked)
        toolbar.insert(editButton, -1)

        delButton = Gtk.ToolButton(stock_id=Gtk.STOCK_REMOVE)
        delButton.set_tooltip_text(_("Remove selected filter"))
        delButton.connect("clicked", self.on_del_clicked)
        toolbar.insert(delButton, -1)

        addButton = Gtk.ToolButton(stock_id=Gtk.STOCK_ADD)
        addButton.set_tooltip_text(_("Add new filter"))
        addButton.connect("clicked", self.on_add_clicked)
        toolbar.insert(addButton, -1)

        addSeqButton = Gtk.ToolButton()
        addSeqButton.set_label(_("Seq"))
        addSeqButton.set_is_important(True)
        addSeqButton.set_tooltip_text(
            _(
                "Create new sequence where listed conditions may be satisfied at different times in a game"
            )
        )
        addSeqButton.connect("clicked", self.on_add_sequence_clicked)
        toolbar.insert(addSeqButton, -1)

        addStreakButton = Gtk.ToolButton()
        addStreakButton.set_label(_("Str"))
        addStreakButton.set_is_important(True)
        addStreakButton.set_tooltip_text(
            _(
                "Create new streak sequence where listed conditions have to be satisfied in consecutive (half)moves"
            )
        )
        addStreakButton.connect("clicked", self.on_add_streak_clicked)
        toolbar.insert(addStreakButton, -1)

        self.filterButton = Gtk.ToggleToolButton(Gtk.STOCK_FIND)
        self.filterButton.set_tooltip_text(_("Filter game list by various conditions"))
        self.filterButton.connect("clicked", self.on_filter_clicked)
        toolbar.insert(self.filterButton, -1)

        tool_box = Gtk.Box()
        tool_box.pack_start(toolbar, False, False, 0)
        self.box.pack_start(tool_box, False, False, 0)

        self.box.show_all()

    def on_filter_clicked(self, button):
        self.filtered = button.get_active()
        if not self.filtered:
            self.persp.preview_panel.filterButton.set_sensitive(True)
            self.persp.opening_tree_panel.filterButton.set_sensitive(True)
            self.clear_filters()
        else:
            self.persp.preview_panel.filterButton.set_sensitive(False)
            self.persp.opening_tree_panel.filterButton.set_sensitive(False)
            self.update_filters()

    def on_del_clicked(self, button):
        selection = self.get_selection()
        model, treeiter = selection.get_selected()
        if treeiter is None:
            return

        self.treestore.remove(treeiter)

        self.update_filters()

    def on_add_sequence_clicked(self, button):
        selection = self.get_selection()
        model, treeiter = selection.get_selected()
        if treeiter is None:
            it = self.treestore.append(None, [_("Sequence"), {}, NONE, SEQUENCE])
            self.get_selection().select_iter(it)
        else:
            text, query, query_type, row_type = self.treestore[treeiter]
            if row_type == RULE:
                it = self.treestore.append(None, [_("Sequence"), {}, NONE, SEQUENCE])
                self.get_selection().select_iter(it)

    def on_add_streak_clicked(self, button):
        selection = self.get_selection()
        model, treeiter = selection.get_selected()
        if treeiter is None:
            it = self.treestore.append(None, [_("Streak"), {}, NONE, STREAK])
            self.get_selection().select_iter(it)
        else:
            text, query, query_type, row_type = self.treestore[treeiter]
            if row_type == RULE:
                it = self.treestore.append(None, [_("Streak"), {}, NONE, STREAK])
                self.get_selection().select_iter(it)
            elif row_type == SEQUENCE:
                it = self.treestore.append(treeiter, [_("Streak"), {}, NONE, STREAK])
                self.get_selection().select_iter(it)
                self.expand_all()

    def on_add_clicked(self, button):
        self.widgets["tag_filter"].set_sensitive(True)
        self.widgets["material_filter"].set_sensitive(True)
        self.widgets["pattern_filter"].set_sensitive(True)

        self.widgets["filter_notebook"].set_current_page(TAG_FILTER)

        self.ini_widgets_from_query({})

        selection = self.get_selection()
        model, treeiter = selection.get_selected()

        if treeiter is not None:
            text, query, query_type, row_type = self.treestore[treeiter]
            if row_type == RULE:
                treeiter = None

        def on_response(dialog, response):
            if response == Gtk.ResponseType.OK:
                (
                    tag_query,
                    material_query,
                    pattern_query,
                ) = self.get_queries_from_widgets()

                if tag_query:
                    it = self.treestore.append(
                        treeiter, [formatted(tag_query), tag_query, TAG_FILTER, RULE]
                    )
                    self.get_selection().select_iter(it)

                if material_query:
                    it = self.treestore.append(
                        treeiter,
                        [
                            formatted(material_query),
                            material_query,
                            MATERIAL_FILTER,
                            RULE,
                        ],
                    )
                    self.get_selection().select_iter(it)

                if pattern_query:
                    it = self.treestore.append(
                        treeiter,
                        [formatted(pattern_query), pattern_query, PATTERN_FILTER, RULE],
                    )
                    self.get_selection().select_iter(it)

                self.expand_all()

                self.update_filters()

                if (not self.filtered) and len(self.treestore) == 1:
                    self.filterButton.set_active(True)

            if hasattr(self, "board_control"):
                self.board_control.emit("action", "CLOSE", None, None)

            self.dialog.hide()
            self.dialog.disconnect(handler_id)

        handler_id = self.dialog.connect("response", on_response)
        self.dialog.show()

    def on_edit_clicked(self, button):
        selection = self.get_selection()
        model, treeiter = selection.get_selected()
        if treeiter is None:
            return

        text, query, query_type, row_type = self.treestore[treeiter]
        if row_type != RULE:
            return

        self.ini_widgets_from_query(query)

        if query_type == TAG_FILTER:
            self.widgets["tag_filter"].set_sensitive(True)
            self.widgets["material_filter"].set_sensitive(False)
            self.widgets["pattern_filter"].set_sensitive(False)
            self.widgets["filter_notebook"].set_current_page(TAG_FILTER)

        elif query_type == MATERIAL_FILTER:
            self.widgets["material_filter"].set_sensitive(True)
            self.widgets["tag_filter"].set_sensitive(False)
            self.widgets["pattern_filter"].set_sensitive(False)
            self.widgets["filter_notebook"].set_current_page(MATERIAL_FILTER)

        elif query_type == PATTERN_FILTER:
            self.widgets["pattern_filter"].set_sensitive(True)
            self.widgets["tag_filter"].set_sensitive(False)
            self.widgets["material_filter"].set_sensitive(False)
            self.widgets["filter_notebook"].set_current_page(PATTERN_FILTER)

        def on_response(dialog, response):
            if response == Gtk.ResponseType.OK:
                (
                    tag_query,
                    material_query,
                    pattern_query,
                ) = self.get_queries_from_widgets()

                if tag_query and query_type == TAG_FILTER:
                    self.treestore[treeiter] = [
                        formatted(tag_query),
                        tag_query,
                        TAG_FILTER,
                        RULE,
                    ]

                if material_query and query_type == MATERIAL_FILTER:
                    self.treestore[treeiter] = [
                        formatted(material_query),
                        material_query,
                        MATERIAL_FILTER,
                        RULE,
                    ]

                if pattern_query and query_type == PATTERN_FILTER:
                    self.treestore[treeiter] = [
                        formatted(pattern_query),
                        pattern_query,
                        PATTERN_FILTER,
                        RULE,
                    ]

                self.update_filters()

            if hasattr(self, "board_control"):
                self.board_control.emit("action", "CLOSE", None, None)

            self.dialog.hide()
            self.dialog.disconnect(handler_id)

        handler_id = self.dialog.connect("response", on_response)
        self.dialog.show()

    def add_sub_fen(self, sub_fen):
        selection = self.get_selection()
        model, treeiter = selection.get_selected()

        if treeiter is not None:
            text, query, query_type, row_type = self.treestore[treeiter]
            if row_type == RULE:
                treeiter = None

        query = {"sub-fen": sub_fen}
        self.treestore.append(treeiter, [formatted(query), query, PATTERN_FILTER, RULE])
        self.expand_all()
        if self.filtered:
            self.update_filters()

    def clear_filters(self):
        self.persp.chessfile.set_tag_filter(None)
        self.persp.chessfile.set_scout_filter(None)
        self.persp.gamelist.load_games()

    def update_filters(self):
        tag_query = {}
        scout_query = {}

        # level 0
        for row in self.treestore:
            text, query, filter_type, row_type = row

            if row_type == RULE:
                if filter_type == TAG_FILTER:
                    tag_query.update(query)
                else:
                    scout_query.update(query)

            elif row_type == SEQUENCE:
                scout_query["sequence"] = []

                # level 1
                for sub_row in row.iterchildren():
                    stext, squery, sfilter_type, srow_type = sub_row

                    if srow_type == RULE:
                        scout_query["sequence"].append(squery)

                    elif srow_type == STREAK:
                        sub_streak = {"streak": []}

                        # level 2
                        for sub_sub_row in sub_row.iterchildren():
                            sstext, ssquery, ssfilter_type, ssrow_type = sub_sub_row
                            if ssrow_type == RULE:
                                sub_streak["streak"].append(ssquery)

                        scout_query["sequence"].append(sub_streak)

            elif row_type == STREAK:
                scout_query["streak"] = []

                # level 1
                for sub_row in row.iterchildren():
                    stext, squery, sfilter_type, srow_type = sub_row

                    if srow_type == RULE:
                        scout_query["streak"].append(squery)

        need_update = False
        if tag_query != self.persp.chessfile.tag_query:
            if self.filtered:
                self.persp.chessfile.set_tag_filter(tag_query)
            need_update = True

        if scout_query != self.persp.chessfile.scout_query:
            if self.filtered:
                self.persp.chessfile.set_scout_filter(scout_query)
            need_update = True

        textbuffer = self.widgets["scout_textbuffer"]
        (iter_first, iter_last) = textbuffer.get_bounds()
        text = textbuffer.get_text(iter_first, iter_last, False)
        if text:
            q = ast.literal_eval(text)
            self.persp.chessfile.set_scout_filter(q)
            need_update = True

        if need_update and self.filtered:
            self.persp.gamelist.load_games()

    def ini_widgets_from_query(self, query):
        """Set filter dialog widget values based on query dict key-value pairs"""

        rule = "variant"
        if rule in query:
            index = 0
            model = self.widgets["variant"].get_model()
            for index, row in enumerate(model):
                if query[rule] == row[1]:
                    break
            self.widgets["variant"].set_active(index)

        for rule in (
            "white",
            "black",
            "event",
            "site",
            "date_from",
            "date_to",
            "eco_from",
            "eco_to",
            "annotator",
        ):
            if rule in query:
                self.widgets[rule].set_text(query[rule])
            else:
                self.widgets[rule].set_text("")

        for rule in ("elo_from", "elo_to"):
            if rule in query:
                self.widgets[rule].set_value(query[rule])
            else:
                self.widgets[rule].set_value(0)

        if "ignore_tag_colors" in query:
            self.widgets["ignore_tag_colors"].set_active(True)
        else:
            self.widgets["ignore_tag_colors"].set_active(False)

        if "result" in query:
            if query["result"] == "1-0":
                self.widgets["result_1_0"].set_active(True)
            elif query["result"] == "0-1":
                self.widgets["result_0_1"].set_active(True)
            elif query["result"] == "1/2-1/2":
                self.widgets["result_1_2"].set_active(True)
            elif query["result"] == "*":
                self.widgets["result_0_0"].set_active(True)
        else:
            self.widgets["result_1_0"].set_active(False)
            self.widgets["result_0_1"].set_active(False)
            self.widgets["result_1_2"].set_active(False)
            self.widgets["result_0_0"].set_active(False)

        q = None
        white0 = ""
        black0 = ""
        if "material" in query:
            q = query["material"]
        if "imbalance" in query:
            q = query["imbalance"]
            self.widgets["imbalance"].set_active(True)
        else:
            self.widgets["imbalance"].set_active(False)

        self.widgets["ignore_material_colors"].set_active(False)
        if type(q) is list and len(q) == 2:
            if "material" in query:
                _, white0, black0 = q[0].split("K")
                _, white1, black1 = q[1].split("K")
            else:
                white0, black0 = q[0].split("v")
                white1, black1 = q[1].split("v")

            if white0 == black1 and black0 == white1:
                self.widgets["ignore_material_colors"].set_active(True)
        elif q is not None:
            if "material" in query:
                _, white0, black0 = q.split("K")
            else:
                white0, black0 = q.split("v")

        for piece in "QRBNP":
            w = white0.count(piece)
            self.widgets["w%s" % piece.lower()].set_value(w if w > 0 else 0)

            b = black0.count(piece)
            self.widgets["b%s" % piece.lower()].set_value(b if b > 0 else 0)

        if "white-move" in query:
            self.widgets["white_move"].set_text(", ".join(query["white-move"]))
        else:
            self.widgets["white_move"].set_text("")

        if "black-move" in query:
            self.widgets["black_move"].set_text(", ".join(query["black-move"]))
        else:
            self.widgets["black_move"].set_text("")

        moved = "moved" in query
        for piece in "pnbrqk":
            active = moved and piece.upper() in query["moved"]
            self.widgets["moved_%s" % piece].set_active(active)

        captured = "captured" in query
        for piece in "pnbrq":
            active = captured and piece.upper() in query["captured"]
            self.widgets["captured_%s" % piece].set_active(active)

        if captured and query["captured"] == "":
            self.widgets["captured_0"].set_active(True)
        else:
            self.widgets["captured_0"].set_active(False)

        if "stm" in query:
            self.widgets["stm"].set_active(True)
            if query["stm"] == "white":
                self.widgets["stm_white"].set_active(True)
            else:
                self.widgets["stm_black"].set_active(True)
        else:
            self.widgets["stm"].set_active(False)

        if "sub-fen" in query:
            sub_fen = query["sub-fen"]
            fen_str = "%s/prnsqkPRNSQK w" % sub_fen
        else:
            sub_fen = ""
            fen_str = "8/8/8/8/8/8/8/8/prnsqkPRNSQK w"
        self.widgets["sub_fen"].set_text(sub_fen)

        # Add a BoardControl widget to dock and initialize it with a new SetupModel
        self.setupmodel = SetupModel()
        self.board_control = BoardControl(self.setupmodel, {}, setup_position=True)
        self.setupmodel.curplayer = SetupPlayer(self.board_control)
        self.setupmodel.connect("game_changed", self.game_changed)

        child = self.widgets["setup_pattern_dock"].get_child()
        if child is not None:
            self.widgets["setup_pattern_dock"].remove(child)
        self.widgets["setup_pattern_dock"].add(self.board_control)
        self.board_control.show_all()

        self.setupmodel.boards = [self.setupmodel.variant(setup=fen_str)]
        self.setupmodel.variations = [self.setupmodel.boards]

        self.setupmodel.start()

        textbuffer = self.widgets["scout_textbuffer"]
        textbuffer.set_text("")

    def get_queries_from_widgets(self):
        """Build tag and scout query dict from filter dialog widget names and values"""

        tag_query = {}
        material_query = {}
        pattern_query = {}

        tree_iter = self.widgets["variant"].get_active_iter()
        if tree_iter is not None:
            model = self.widgets["variant"].get_model()
            variant_code = model[tree_iter][1]
            tag_query["variant"] = variant_code

        for rule in (
            "white",
            "black",
            "event",
            "site",
            "date_from",
            "date_to",
            "eco_from",
            "eco_to",
            "annotator",
        ):
            if self.widgets[rule].get_text():
                tag_query[rule] = self.widgets[rule].get_text()

        for rule in ("elo_from", "elo_to"):
            if self.widgets[rule].get_value_as_int():
                tag_query[rule] = self.widgets[rule].get_value_as_int()

        if self.widgets["ignore_tag_colors"].get_active():
            tag_query["ignore_tag_colors"] = True

        if self.widgets["result_1_0"].get_active():
            tag_query["result"] = "1-0"
        if self.widgets["result_0_1"].get_active():
            tag_query["result"] = "0-1"
        if self.widgets["result_1_2"].get_active():
            tag_query["result"] = "1/2-1/2"
        if self.widgets["result_0_0"].get_active():
            tag_query["result"] = "*"

        w_material = []
        for piece in "qrbnp":
            w_material.append(
                piece.upper() * self.widgets["w%s" % piece].get_value_as_int()
            )

        b_material = []
        for piece in "qrbnp":
            b_material.append(
                piece.upper() * self.widgets["b%s" % piece].get_value_as_int()
            )

        w_material = "".join(w_material)
        b_material = "".join(b_material)

        if w_material or b_material:
            if self.widgets["imbalance"].get_active():
                material_query["imbalance"] = f"{w_material}v{b_material}"

                if self.widgets["ignore_material_colors"].get_active():
                    material_query["imbalance"] = [
                        f"{w_material}v{b_material}",
                        f"{b_material}v{w_material}",
                    ]
            else:
                material_query["material"] = f"K{w_material}K{b_material}"

                if self.widgets["ignore_material_colors"].get_active():
                    material_query["material"] = [
                        f"K{w_material}K{b_material}",
                        f"K{b_material}K{w_material}",
                    ]

        if self.widgets["white_move"].get_text():
            moves = [
                move.strip()
                for move in self.widgets["white_move"].get_text().split(",")
            ]
            material_query["white-move"] = moves

        if self.widgets["black_move"].get_text():
            moves = [
                move.strip()
                for move in self.widgets["black_move"].get_text().split(",")
            ]
            material_query["black-move"] = moves

        moved = ""
        for piece in "pnbrqk":
            if self.widgets["moved_%s" % piece].get_active():
                moved += piece.upper()
        if moved:
            material_query["moved"] = "%s" % moved

        captured = ""
        for piece in "pnbrq0":
            if self.widgets["captured_%s" % piece].get_active():
                captured += piece.upper()
        if captured:
            material_query["captured"] = "%s" % captured.replace("0", "")

        if self.widgets["stm"].get_active():
            if self.widgets["stm_white"].get_active():
                material_query["stm"] = "white"
            else:
                material_query["stm"] = "black"

        if self.widgets["sub_fen"].get_text():
            pattern_query["sub-fen"] = self.widgets["sub_fen"].get_text()

        return (tag_query, material_query, pattern_query)

    def fen_changed(self):
        self.widgets["sub_fen"].set_text(self.get_fen())

    def on_copy_sub_fen(self, widget):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        text = self.widgets["sub_fen"].get_text()
        if len(text) > 0:
            clipboard.set_text(text, -1)

    def on_paste_sub_fen(self, widget):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        text = clipboard.wait_for_text()
        if text.count("/") == 7:
            self.board_control.emit("action", "SETUP", None, text)

    def game_changed(self, model, ply):
        GLib.idle_add(self.fen_changed)

    def get_fen(self):
        return self.setupmodel.boards[-1].as_fen(NORMALCHESS)



"""Upside-down Chess variant"""

from pychess.Utils.const import UPSIDEDOWNCHESS, VARIANTS_OTHER
from pychess.Utils.Board import Board

UPSIDEDOWNSTART = "RNBQKBNR/PPPPPPPP/8/8/8/8/pppppppp/rnbqkbnr w - - 0 1"


class UpsideDownBoard(Board):
    variant = UPSIDEDOWNCHESS
    __desc__ = _(
        "FICS wild/5: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "http://en.wikipedia.org/wiki/Chess_variant#Chess_with_different_starting_positions\n"
        + "Pawns start on their 7th rank rather than their 2nd rank!"
    )
    name = _("Upside Down")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_OTHER

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=UPSIDEDOWNSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import builtins

from .const import (
    ADJOURNED_LOST_CONNECTION_BLACK,
    UNKNOWN_REASON,
    WON_LESSMATERIAL,
    KILLED,
    ADJOURNED_SERVER_SHUTDOWN,
    DRAW_EQUALMATERIAL,
    ABORTED_ADJUDICATION,
    DISCONNECTED,
    BLACKWON,
    WHITEWON,
    DRAW_CALLFLAG,
    WON_RESIGN,
    DRAW,
    WON_CALLFLAG,
    DRAW_WHITEINSUFFICIENTANDBLACKTIME,
    DRAW_LENGTH,
    WON_MATE,
    WON_KINGINEIGHTROW,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    ADJOURNED_COURTESY_WHITE,
    WON_THREECHECK,
    WON_WIPEOUT,
    DRAW_AGREE,
    DRAW_INSUFFICIENT,
    DRAW_STALEMATE,
    DRAW_ADJUDICATION,
    UNKNOWN_STATE,
    ADJOURNED_LOST_CONNECTION,
    WON_NOMATERIAL,
    ADJOURNED,
    WON_DISCONNECTION,
    ABORTED,
    ADJOURNED_AGREEMENT,
    WON_KINGEXPLODE,
    WON_KINGINCENTER,
    ABORTED_SERVER_SHUTDOWN,
    ABORTED_AGREEMENT,
    WHITE_ENGINE_DIED,
    BLACK_ENGINE_DIED,
    WON_ADJUDICATION,
    ADJOURNED_COURTESY_BLACK,
    ADJOURNED_COURTESY,
    DRAW_REPETITION,
    ABORTED_COURTESY,
    DRAW_50MOVES,
    ADJOURNED_LOST_CONNECTION_WHITE,
    ABORTED_EARLY,
    ABORTED_DISCONNECTION,
    CANCELLED,
    PRACTICE_GOAL_REACHED,
    DRAW_KINGSINEIGHTROW,
)


if "_" not in builtins.__dict__:
    builtins.__dict__["_"] = lambda s: s
    builtins.__dict__["ngettext"] = lambda singular, plural, n: (
        singular if n == 1 else plural
    )

reprColor = [_("White"), _("Black")]

reprPiece = [
    "Empty",
    _("Pawn"),
    _("Knight"),
    _("Bishop"),
    _("Rook"),
    _("Queen"),
    _("King"),
    _("Hawk"),
    _("Elephant"),
    "BPawn",
]

localReprSign = ["", _("P"), _("N"), _("B"), _("R"), _("Q"), _("K"), _("H"), _("E")]

reprResult_long = {
    DRAW: _("The game ended in a draw"),
    WHITEWON: _("%(white)s won the game"),
    BLACKWON: _("%(black)s won the game"),
    KILLED: _("The game has been killed"),
    ADJOURNED: _("The game has been adjourned"),
    ABORTED: _("The game has been aborted"),
    UNKNOWN_STATE: _("Unknown game state"),
    CANCELLED: _("Game cancelled"),
}

reprReason_long = {
    DRAW_INSUFFICIENT: _("Because neither player has sufficient material to mate"),
    DRAW_REPETITION: _("Because the same position was repeated three times in a row"),
    DRAW_50MOVES: _("Because the last 50 moves brought nothing new"),
    DRAW_CALLFLAG: _("Because both players ran out of time"),
    DRAW_STALEMATE: _("Because %(mover)s stalemated"),
    DRAW_AGREE: _("Because both players agreed to a draw"),
    DRAW_ADJUDICATION: _("Because of adjudication by an admin"),
    DRAW_LENGTH: _("Because the game exceed the max length"),
    DRAW_BLACKINSUFFICIENTANDWHITETIME: _(
        "Because %(white)s ran out of time and %(black)s has insufficient material to mate"
    ),
    DRAW_WHITEINSUFFICIENTANDBLACKTIME: _(
        "Because %(black)s ran out of time and %(white)s has insufficient material to mate"
    ),
    DRAW_EQUALMATERIAL: _("Because both players have the same amount of pieces"),
    DRAW_KINGSINEIGHTROW: _("Because both king reached the eight row"),
    WON_RESIGN: _("Because %(loser)s resigned"),
    WON_CALLFLAG: _("Because %(loser)s ran out of time"),
    WON_MATE: _("Because %(loser)s was checkmated"),
    WON_DISCONNECTION: _("Because %(loser)s disconnected"),
    WON_ADJUDICATION: _("Because of adjudication by an admin"),
    WON_LESSMATERIAL: _("Because %(winner)s has fewer pieces"),
    WON_NOMATERIAL: _("Because %(winner)s lost all pieces"),
    WON_KINGEXPLODE: _("Because %(loser)s king exploded"),
    WON_KINGINCENTER: _("Because %(winner)s king reached the center"),
    WON_THREECHECK: _("Because %(winner)s was giving check 3 times"),
    WON_KINGINEIGHTROW: _("Because %(winner)s king reached the eight row"),
    WON_WIPEOUT: _("Because %(winner)s wiped out white horde"),
    ADJOURNED_LOST_CONNECTION: _("Because a player lost connection"),
    ADJOURNED_AGREEMENT: _("Because both players agreed to an adjournment"),
    ADJOURNED_SERVER_SHUTDOWN: _("Because the server was shut down"),
    ADJOURNED_COURTESY: _(
        "Because a player lost connection and the other player requested adjournment"
    ),
    ADJOURNED_COURTESY_WHITE: _(
        "Because %(black)s lost connection to the server and %(white)s requested adjournment"
    ),
    ADJOURNED_COURTESY_BLACK: _(
        "Because %(white)s lost connection to the server and %(black)s requested adjournment"
    ),
    ADJOURNED_LOST_CONNECTION_WHITE: _(
        "Because %(white)s lost connection to the server"
    ),
    ADJOURNED_LOST_CONNECTION_BLACK: _(
        "Because %(black)s lost connection to the server"
    ),
    ABORTED_ADJUDICATION: _(
        "Because of adjudication by an admin. No rating changes have occurred."
    ),
    ABORTED_AGREEMENT: _(
        "Because both players agreed to abort the game. No rating changes have occurred."
    ),
    ABORTED_COURTESY: _(
        "Because of courtesy by a player. No rating changes have occurred."
    ),
    ABORTED_EARLY: _(
        "Because a player aborted the game. Either player can abort the game without \
      the other's consent before the second move. No rating changes have occurred."
    ),
    ABORTED_DISCONNECTION: _(
        "Because a player disconnected and there are too few moves to warrant adjournment. No rating changes have occurred."
    ),
    ABORTED_SERVER_SHUTDOWN: _(
        "Because the server was shut down. No rating changes have occurred."
    ),
    WHITE_ENGINE_DIED: _("Because the %(white)s engine died"),
    BLACK_ENGINE_DIED: _("Because the %(black)s engine died"),
    DISCONNECTED: _("Because the connection to the server was lost"),
    UNKNOWN_REASON: _("The reason is unknown"),
    PRACTICE_GOAL_REACHED: _("Because practice goal reached"),
}



import time
from threading import Thread
from pychess.System import fident


def repeat(func, *args, **kwargs):
    """Repeats a function in a new thread until it returns False"""

    def run():
        while func(*args, **kwargs):
            pass

    thread = Thread(target=run, name=fident(func))
    thread.daemon = True
    thread.start()


def repeat_sleep(func, sleeptime, recur=False):
    """
    Runs func in a thread and repeats it approximately each sleeptime [s]
    until func returns False. Notice that we sleep first, then run. Not the
    other way around. If repeat_sleep is called with recur=True, each call
    will be called with the return value of last call as argument. The
    argument has to be optional, as it wont be used first time, and it has
    to be non-None.
    """

    def run():
        last = time.time()
        val = None
        while True:
            time.sleep(time.time() - last + sleeptime)
            if not time:
                # If python has been shutdown while we were sleeping, the
                # imported modules will be None
                return
            last = time.time()
            if recur and val:
                val = func(val)
            else:
                val = func()
            if not val:
                break

    thread = Thread(target=run, name=fident(func))
    thread.daemon = True
    thread.start()



"""Wildcastle shuffle Chess"""

import random

from pychess.Utils.const import WILDCASTLESHUFFLECHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board


class WildcastleShuffleBoard(Board):
    variant = WILDCASTLESHUFFLECHESS
    __desc__ = _(
        "xboard wildcastle http://home.hccnet.nl/h.g.muller/engine-intf.html#8\n"
        + "FICS wild/1: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "* In this variant both sides have the same set of pieces as in normal chess.\n"
        + "* The white king starts on d1 or e1 and the black king starts on d8 or e8,\n"
        + "* and the rooks are in their usual positions.\n"
        + "* Bishops are always on opposite colors.\n"
        + "* Subject to these constraints the position of the \
                 pieces on their first ranks is random.\n"
        + "* Castling is done similarly to normal chess:\n"
        + "* o-o-o indicates long castling and o-o short castling."
    )
    name = _("Wildcastle shuffle")
    cecp_name = "wildcastle"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.shuffle_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def shuffle_start(self):
        def get_shuffle():
            positions = [2, 3, 4, 5, 6, 7]
            back_rank = ["r"] + ([""] * 6) + ["r"]

            king = random.choice((4, 5))
            back_rank[king - 1] = "k"
            positions.remove(king)

            bishop = random.choice(positions)
            back_rank[bishop - 1] = "b"
            positions.remove(bishop)
            color = bishop % 2

            bishop = random.choice([i for i in positions if i % 2 != color])
            back_rank[bishop - 1] = "b"
            positions.remove(bishop)

            queen = random.choice(positions)
            back_rank[queen - 1] = "q"
            positions.remove(queen)

            knight = random.choice(positions)
            back_rank[knight - 1] = "n"
            positions.remove(knight)

            knight = random.choice(positions)
            back_rank[knight - 1] = "n"
            positions.remove(knight)

            return "".join(back_rank)

        fen = (
            get_shuffle()
            + "/pppppppp/8/8/8/8/PPPPPPPP/"
            + get_shuffle().upper()
            + " w KQkq - 0 1"
        )

        return fen



from .lutils.attack import getAttacks, piecesAttackingCord
from .const import NB_OF_CASES
from .lutils.ldata import PIECE_VALUES
from .Cord import Cord
from pychess.Utils.const import KING


# This file contains algorithm helping the users to choose the right decisions
# The first algorithm gives to the user all the pieces attacked and not protected by the foe
class DecisionSupportAlgorithm:
    def __init__(self):
        self.local_game = False
        self.activationEnabled = False

        # to avoid calculating once more the coordinates, we save them
        self.coordinate_in_danger = []

    def enableDisableAlgo(self, enable_algorithm):
        self.activationEnabled = enable_algorithm

    def set_foe_as_bot(self):
        self.local_game = True

    def calculate_coordinate_in_danger(self, board, mycolor, newTurn=True):
        """this function should be used for applying the algorithm"""
        if not newTurn:
            return self.coordinate_in_danger

        self.coordinate_in_danger = []

        self.coordinate_in_danger = self.__apply_algorithm(
            board, mycolor, self.__not_protected
        )
        self.coordinate_in_danger += self.__apply_algorithm(
            board, mycolor, self.__attacked_and_not_protected
        )

        return self.coordinate_in_danger

    def __apply_algorithm(self, board, mycolor, algorithm):
        """returns the list of coordinate in danger for player playing with mycolor pieces
         according to the algorithm used
        args : board of type Board, coordinate of type Cord, algorithm : function taking two arguments
        board, mycolor and the current coordinate we are looking

        WARNING : The king is currently excluded of the calculus => it is never considered as in danger
        """
        coordinate_in_danger = []

        if self.local_game and self.activationEnabled:
            # TODO: tests
            for i in range(NB_OF_CASES):
                c = Cord(i)
                number, letter = c.cords
                piece = board.data[letter][number]
                if piece is not None and piece.piece is not KING:
                    if piece.color == mycolor:
                        coordinate_in_color = algorithm(board, mycolor, c)
                        if coordinate_in_color is not None:
                            coordinate_in_danger.append(coordinate_in_color)

        return coordinate_in_danger

    def __not_protected(self, board, mycolor, coordinate):
        """returns the coord with the color yellow if the coord is not protected, None if it is
        args : board of type Board, coordinate of type Cord, mycolor of value BLACK or WHITE
        which is the color of the player who will play the next move"""

        c_colored = None
        attacks = getAttacks(board.board, coordinate.cord, 1 - mycolor)
        defense = getAttacks(board.board, coordinate.cord, mycolor)

        # if one of the piece is not protected but not attacked, danger
        # means that towers are highlighted at the start of the game
        if attacks == 0 and defense == 0:
            c_colored = Cord(coordinate.cord, color="Y")

        return c_colored

    def __attacked_and_not_protected(self, board, mycolor, coordinate):
        """returns the coord with the color red if the coord is not protected and attacked, None if it is not
        args : board of type Board, cord of type Cord, mycolor of value BLACK or WHITE
        which is the color of the player who will play the next move"""

        pieces_attacking = piecesAttackingCord(
            board.board, coordinate.cord, 1 - mycolor
        )
        defense = getAttacks(board.board, coordinate.cord, mycolor)

        c_colored = None

        number, letter = coordinate.cords
        piece = board.data[letter][number]

        if len(pieces_attacking) != 0:
            # if one of the piece is not protected and attacked, very dangerous
            if defense == 0:
                c_colored = Cord(coordinate.cord, color="R")
            else:
                # if one of the piece attacking is weaker than piece attacked, very dangerous

                # in this part we find the weakest ennemy pieces attacking our piece
                piece_attacking = pieces_attacking[0]
                min_value = PIECE_VALUES[piece_attacking]
                for piece_found in pieces_attacking:
                    if PIECE_VALUES[piece_found] <= min_value:
                        min_value = PIECE_VALUES[piece_found]
                        piece_attacking = piece_found

                # then we compare it to the value of the piece
                # currently, we consider trades as not dangerous, this behaviour could be changed
                if PIECE_VALUES[piece_attacking] < PIECE_VALUES[piece.piece]:
                    c_colored = Cord(coordinate.cord, color="R")

        return c_colored



import asyncio
import re
import socket
import webbrowser
from collections import defaultdict

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import GObject

from pychess.System import uistuff, conf
from pychess.widgets import mainwindow
from pychess.ic.FICSConnection import (
    FICSMainConnection,
    FICSHelperConnection,
    LogOnException,
)
from pychess.perspectives import perspective_manager

host = None
port = None

dialog = None


def run():
    global dialog
    dialog.widgets["fics_logon"].set_transient_for(mainwindow())

    if dialog.lounge:
        dialog.lounge.present()
    else:
        dialog.present()


def stop():
    global dialog
    if dialog is not None:
        dialog._disconnect()


class AutoLogoutException(Exception):
    pass


class ICLogon:
    def __init__(self):
        self.connection = None
        self.helperconn = None
        self.connection_task = None
        self.helperconn_task = None
        self.lounge = None
        self.canceled = False
        self.cids = defaultdict(list)
        self.widgets = uistuff.GladeWidgets("fics_logon.glade")
        uistuff.keepWindowSize(
            "fics_logon",
            self.widgets["fics_logon"],
            defaultPosition=uistuff.POSITION_GOLDEN,
        )
        self.widgets["fics_logon"].connect(
            "key-press-event", lambda w, e: e.keyval == Gdk.KEY_Escape and w.hide()
        )

        self.ics = "FICS"
        self.as_guest = self.widgets["logOnAsGuest"]

        self.widgets["logOnAsGuest"].connect("toggled", self.on_logOnAsGuest_toggled)

        def on_username_changed(widget):
            conf.set(
                "usernameEntry", self.user_name_get_value(widget), section=self.ics
            )

        self.widgets["nameEntry"].connect("changed", on_username_changed)

        def on_password_changed(widget):
            conf.set("passwordEntry", widget.get_text(), section=self.ics)

        self.widgets["passEntry"].connect("changed", on_password_changed)

        def on_host_changed(widget):
            conf.set("hostEntry", self.host_get_value(widget), section=self.ics)

        self.widgets["hostEntry"].connect("changed", on_host_changed)

        self.widgets["timesealCheck"].connect("toggled", self.on_timeseal_toggled)

        self.infobar = Gtk.InfoBar()
        self.infobar.set_message_type(Gtk.MessageType.WARNING)
        # self.widgets["messagePanelHBox"].pack_start(self.infobar,
        #    expand=False, fill=False)
        self.widgets["messagePanelHBox"].pack_start(self.infobar, False, False, 0)
        self.widgets["cancelButton"].connect("clicked", self.onCancel, True)
        self.widgets["stopButton"].connect("clicked", self.onCancel, False)
        self.widgets["createNewButton"].connect("clicked", self.onCreateNew)
        self.widgets["connectButton"].connect("clicked", self.onConnectClicked)

        self.widgets["progressbar"].set_show_text(True)

    def get_user_names(self, value=None):
        """Split and return usernameEntry config item into registered and guest username"""
        if value is not None:
            names = value.split("|")
        else:
            names = conf.get("usernameEntry", section=self.ics).split("|")
        if len(names) == 0:
            names = ["", ""]
        elif len(names) < 2:
            names.append(names[0])
        return names

    def user_name_get_value(self, entry):
        names = self.get_user_names()
        if self.as_guest.get_active():
            text = f"{names[0]}|{entry.get_text()}"
        else:
            text = f"{entry.get_text()}|{names[1]}"
        return text

    def user_name_set_value(self, entry, value):
        names = self.get_user_names(value=value)
        if self.as_guest.get_active():
            entry.set_text(names[1])
        else:
            entry.set_text(names[0])

    # workaround to FICS Password input doesnt handle strings starting with a number
    # https://github.com/pychess/pychess/issues/1375
    def password_set_value(self, entry, value):
        entry.set_text(str(value))

    # workaround to Can't type IP to FICS login dialog
    # https://github.com/pychess/pychess/issues/1360
    def host_get_value(self, entry):
        return entry.get_text().replace(".", "|")

    def host_set_value(self, entry, value):
        entry.set_text(str(value).replace("|", "."))

    def on_logOnAsGuest_toggled(self, widget):
        names = self.get_user_names()
        self.widgets["nameEntry"].set_text(
            names[1] if widget.get_active() else names[0]
        )
        if self.ics == "ICC":
            self.widgets["nameLabel"].set_sensitive(not widget.get_active())
            self.widgets["nameEntry"].set_sensitive(not widget.get_active())
        else:
            self.widgets["nameLabel"].set_sensitive(True)
            self.widgets["nameEntry"].set_sensitive(True)
        self.widgets["passwordLabel"].set_sensitive(not widget.get_active())
        self.widgets["passEntry"].set_sensitive(not widget.get_active())
        conf.set("asGuestCheck", widget.get_active(), section=self.ics)

    def on_timeseal_toggled(self, widget):
        conf.set("timesealCheck", widget.get_active(), section=self.ics)

    def on_ics_combo_changed(self, combo):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            is_enabled = model[tree_iter][1]  # Check the 'enabled' column
            if not is_enabled:
                combo.set_active(0)
                return

            self.ics = model[tree_iter][0]
            # print("Selected: %s" % self.ics)
            self.widgets["logOnAsGuest"].set_active(
                conf.get("asGuestCheck", section=self.ics)
            )
            self.on_logOnAsGuest_toggled(self.widgets["logOnAsGuest"])
            self.user_name_set_value(
                self.widgets["nameEntry"], conf.get("usernameEntry", section=self.ics)
            )
            self.password_set_value(
                self.widgets["passEntry"], conf.get("passwordEntry", section=self.ics)
            )
            self.host_set_value(
                self.widgets["hostEntry"], conf.get("hostEntry", section=self.ics)
            )
            self.widgets["timesealCheck"].set_active(
                conf.get("timesealCheck", section=self.ics)
            )
            self.on_timeseal_toggled(self.widgets["timesealCheck"])

    def _disconnect(self):
        for obj in self.cids:
            for cid in self.cids[obj]:
                if obj.handler_is_connected(cid):
                    obj.disconnect(cid)
        self.cids.clear()

        if self.connection is not None:
            self.connection.close()
            if not self.connection_task.cancelled():
                self.connection_task.cancel()
            self.connection = None
        if self.helperconn is not None:
            self.helperconn.close()
            if not self.helperconn_task.cancelled():
                self.helperconn_task.cancel()
            self.helperconn = None
        self.lounge = None

    def _cancel(self):
        if self.connection is not None:
            self.connection.cancel()
        if self.helperconn is not None:
            self.helperconn.cancel()
        self._disconnect()

    def show(self):
        self.widgets["fics_logon"].show()

    def present(self):
        self.widgets["fics_logon"].present()

    def hide(self):
        self.widgets["fics_logon"].hide()

    def showConnecting(self):
        self.widgets["progressbarBox"].show()
        self.widgets["mainbox"].set_sensitive(False)
        self.widgets["connectButton"].hide()
        self.widgets["stopButton"].show()

        def pulse():
            self.widgets["progressbar"].pulse()
            if self.connection is None:
                return False
            else:
                return not self.connection.isConnected()

        self.pulser = GLib.timeout_add(30, pulse)

    def showNormal(self):
        self.widgets["mainbox"].set_sensitive(True)
        self.widgets["connectButton"].show()
        self.widgets["fics_logon"].set_default(self.widgets["connectButton"])
        self.widgets["stopButton"].hide()
        self.widgets["progressbarBox"].hide()
        self.widgets["progressbar"].set_text("")
        GObject.source_remove(self.pulser)

    def showMessage(self, connection, message):
        self.widgets["progressbar"].set_text(message)

    def showError(self, connection, error):
        text = str(error)
        if isinstance(error, IOError):
            title = _("Connection Error")
        elif isinstance(error, LogOnException):
            title = _("Log on Error")
        elif isinstance(error, EOFError):
            title = _("Connection was closed")
        elif isinstance(error, socket.error):
            title = _("Connection Error")
            text = ", ".join(map(str, error.args))
        elif isinstance(error, socket.gaierror) or isinstance(error, socket.herror):
            title = _("Address Error")
            text = ", ".join(map(str, error.args))
        elif isinstance(error, AutoLogoutException):
            title = _("Auto-logout")
            text = _(
                "You have been logged out because you were idle more than 60 minutes"
            )
        else:
            title = str(error.__class__)

        self.showNormal()

        content_area = self.infobar.get_content_area()
        for widget in content_area:
            content_area.remove(widget)
        content = Gtk.HBox()
        image = Gtk.Image()
        image.set_from_stock(Gtk.STOCK_DIALOG_WARNING, Gtk.IconSize.DIALOG)
        content.pack_start(image, False, False, 0)
        vbox = Gtk.VBox()
        label = Gtk.Label()
        label.props.xalign = 0
        label.props.justify = Gtk.Justification.LEFT
        label.set_markup("<b><big>%s</big></b>" % title)
        vbox.pack_start(label, True, False, 0)
        for line in str(text).split("\n"):
            label = Gtk.Label()
            label.set_size_request(476, -1)
            label.props.selectable = True
            label.props.wrap = True
            label.props.xalign = 0
            label.props.justify = Gtk.Justification.LEFT
            label.set_markup(line)
            vbox.pack_start(label, True, False, 0)
        content.pack_start(vbox, False, False, 7)
        content_area.add(content)
        self.widgets["messagePanel"].show_all()

    def onCreateNew(self, button):
        if self.widgets["hostEntry"].get_text() == "chessclub.com":
            webbrowser.open("https://store.chessclub.com/customer/account/create/")
        else:
            webbrowser.open("http://www.freechess.org/Register/index.html")

    def onConnectClicked(self, button):
        self.canceled = False
        self.widgets["messagePanel"].hide()

        if self.widgets["logOnAsGuest"].get_active():
            username = self.widgets["nameEntry"].get_text()
            password = ""
        else:
            username = self.widgets["nameEntry"].get_text()
            password = self.widgets["passEntry"].get_text()

        if port:
            ports = (port,)
        else:
            ports = self.widgets["portsEntry"].get_text()
            ports = list(map(int, re.findall(r"\d+", ports)))
            if 5000 not in ports:
                ports.append(5000)
            if 23 not in ports:
                ports.append(23)
        alternate_host = self.widgets["hostEntry"].get_text()

        timeseal = self.widgets["timesealCheck"].get_active()

        self.showConnecting()
        self.host = (
            host
            if host is not None
            else alternate_host if alternate_host else "freechess.org"
        )
        self.connection = FICSMainConnection(
            self.host, ports, timeseal, username, password
        )
        for signal, callback in (
            ("connected", self.onConnected),
            ("error", self.onConnectionError),
            ("connectingMsg", self.showMessage),
        ):
            self.cids[self.connection].append(self.connection.connect(signal, callback))
        self.main_connected_event = asyncio.Event()
        self.connection_task = asyncio.create_task(self.connection.start())

        # guest users are rather limited on ICC (helper connection is useless)
        if self.host not in ("localhost", "127.0.0.1", "chessclub.com"):
            self.helperconn = FICSHelperConnection(
                self.connection, self.host, ports, timeseal
            )
            self.helperconn.connect("error", self.onHelperConnectionError)

            async def coro():
                await self.main_connected_event.wait()
                await self.helperconn.start()

            self.helperconn_task = asyncio.create_task(coro())

    def onHelperConnectionError(self, connection, error):
        if self.helperconn is not None:
            dialog = Gtk.MessageDialog(
                mainwindow(),
                type=Gtk.MessageType.QUESTION,
                buttons=Gtk.ButtonsType.YES_NO,
            )
            dialog.set_markup(_("Guest logins disabled by FICS server"))
            text = "PyChess can maintain users status and games list only if it changes\n\
            'open', 'gin' and 'availinfo' user variables.\n\
            Do you enable to set these variables on?"
            dialog.format_secondary_text(text)
            response = dialog.run()
            dialog.destroy()

            self.helperconn.cancel()
            self.helperconn.close()
            self.helperconn = None

            set_user_vars = response == Gtk.ResponseType.YES

            async def coro():
                await self.main_connected_event.wait()
                self.connection.start_helper_manager(set_user_vars)

            asyncio.create_task(coro())

    def onConnected(self, connection):
        self.main_connected_event.set()
        if connection.ICC:
            self.connection.start_helper_manager(True)

        self.lounge = perspective_manager.get_perspective("fics")
        self.lounge.open_lounge(connection, self.helperconn, self.host)
        self.hide()
        self.lounge.show()
        self.lounge.connect("logout", lambda iclounge: self.onLogout(connection))
        self.cids[self.lounge].append(
            self.lounge.connect(
                "autoLogout", lambda lounge: self.onAutologout(connection)
            )
        )

        self.showNormal()
        self.widgets["messagePanel"].hide()

    def onCancel(self, widget, hide):
        self.canceled = True

        if self.connection and self.connection.isConnecting():
            self._cancel()
            self.showNormal()
        if hide:
            self.widgets["fics_logon"].hide()

    def onConnectionError(self, connection, error):
        self._disconnect()
        if not self.canceled:
            self.showError(connection, error)
            self.present()

    def onLogout(self, connection):
        self._disconnect()

    def onAutologout(self, connection):
        self._disconnect()
        self.showError(connection, AutoLogoutException())
        self.present()



from .bitboard import bitPosArray, notBitPosArray, lastBit, firstBit, clearBit, lsb
from .ldata import moveArray, rays, directions, fromToRay, PIECE_VALUES, PAWN_VALUE
from pychess.Utils.const import (
    ASEAN_VARIANTS,
    ASEAN_BBISHOP,
    ASEAN_WBISHOP,
    ASEAN_QUEEN,
    SCHESS,
    BLACK,
    WHITE,
    PAWN,
    BPAWN,
    KNIGHT,
    BISHOP,
    ROOK,
    QUEEN,
    KING,
    ENPASSANT,
    ATOMICCHESS,
    HAWK,
    ELEPHANT,
)

#
# Caveat: Many functions in this module has very similar code. If you fix a
# bug, or write a performance enhancement, please update all functions. Apologies
# for the inconvenience
#


def isAttacked(board, cord, color, ischecked=False):
    """To determine if cord is attacked by any pieces from color."""

    _moveArray = moveArray
    pboards = board.boards[color]

    # Knights, Hawks, Elephants
    if (pboards[KNIGHT] | pboards[HAWK] | pboards[ELEPHANT]) & _moveArray[KNIGHT][cord]:
        return True

    rayto = fromToRay[cord]
    blocker = board.blocker

    # Bishops & Queens
    if board.variant in ASEAN_VARIANTS:
        bishopMoves = _moveArray[ASEAN_BBISHOP if color == WHITE else ASEAN_WBISHOP]
        if pboards[BISHOP] & bishopMoves[cord]:
            return True
        if pboards[QUEEN] & _moveArray[ASEAN_QUEEN][cord]:
            return True
    else:
        bitboard = (pboards[BISHOP] | pboards[HAWK] | pboards[QUEEN]) & _moveArray[
            BISHOP
        ][cord]
        if bitboard:
            others = ~bitboard & blocker
            # inlined iterBits()
            while bitboard:
                bit = bitboard & -bitboard
                ray = rayto[lsb[bit]]
                # If there is a path and no other piece stand in our way
                if ray and not ray & others:
                    return True
                bitboard -= bit

    # Rooks & Queens
    if board.variant in ASEAN_VARIANTS:
        bitboard = pboards[ROOK] & _moveArray[ROOK][cord]
    else:
        bitboard = (pboards[ROOK] | pboards[QUEEN] | pboards[ELEPHANT]) & _moveArray[
            ROOK
        ][cord]
    if bitboard:
        others = ~bitboard & blocker
        # inlined iterBits()
        while bitboard:
            bit = bitboard & -bitboard
            ray = rayto[lsb[bit]]
            # If there is a path and no other piece stand in our way
            if ray and not ray & others:
                return True
            bitboard -= bit

            # Pawns
            # Would a pawn of the opposite color, standing at out kings cord, be able
            # to attack any of our pawns?
    ptype = color == WHITE and BPAWN or PAWN
    if pboards[PAWN] & _moveArray[ptype][cord]:
        return True

    # King
    if pboards[KING] & _moveArray[KING][cord]:
        if board.variant == ATOMICCHESS and ischecked:
            return False
        else:
            return True

    return False


# TODO tests
def propagateRayFollowingMovement(board, cord, bitboard):
    "tests if there are pieces blocking the way on the movement line"
    rayto = fromToRay[cord]
    blocker = board.blocker

    bits = 0

    while bitboard:
        bit = bitboard & -bitboard
        c = lsb[bit]
        ray = rayto[c]
        if ray and not clearBit(ray & blocker, c):
            bits |= bitPosArray[c]
        bitboard -= bit

    return bits


def piecesAttackingCord(board, cord, color):
    """return the type of piece attacking Cord
    does not support variant yet
    The type of args are LBoard, Cord as flat number (cord argument of class Cord, BLACK or WHITE from const file
    """

    _moveArray = moveArray
    pieces = board.boards[color]

    pieces_attacking = []

    # Knights
    bits_knight = pieces[KNIGHT] & _moveArray[KNIGHT][cord]
    if bits_knight:
        pieces_attacking.append(KNIGHT)

    # Kings
    bits_king = pieces[KING] & _moveArray[KING][cord]
    if bits_king:
        pieces_attacking.append(KING)

    # Pawns, to test , bug possible with BPAWN
    bits_pawn = pieces[PAWN] & _moveArray[color == WHITE and BPAWN or PAWN][cord]
    if bits_pawn:
        pieces_attacking.append(PAWN)

    # Bishops
    bitboard_bishop = (pieces[BISHOP] | pieces[QUEEN]) & _moveArray[BISHOP][cord]

    if propagateRayFollowingMovement(board, cord, bitboard_bishop):
        pieces_attacking.append(BISHOP)

    bitboard_rook = (pieces[ROOK] | pieces[QUEEN]) & _moveArray[ROOK][cord]
    # inlined iterBits()

    if propagateRayFollowingMovement(board, cord, bitboard_rook):
        pieces_attacking.append(ROOK)

    bitboard_queen_diago = pieces[QUEEN] & _moveArray[BISHOP][cord]
    if propagateRayFollowingMovement(board, cord, bitboard_queen_diago):
        pieces_attacking.append(QUEEN)

    bitboard_queen_straight = pieces[QUEEN] & _moveArray[ROOK][cord]
    if propagateRayFollowingMovement(board, cord, bitboard_queen_straight):
        pieces_attacking.append(QUEEN)

    return pieces_attacking


def getAttacks(board, cord, color):
    """To create a bitboard of pieces of color, which attacks cord
    The type of args are LBoard, ,BLACK or WHITE from const file
    """

    _moveArray = moveArray
    pieces = board.boards[color]

    # Knights
    bits = (pieces[KNIGHT] | pieces[HAWK] | pieces[ELEPHANT]) & _moveArray[KNIGHT][cord]

    # Kings
    bits |= pieces[KING] & _moveArray[KING][cord]

    # Pawns, to test , bug possible with BPAWN
    bits |= pieces[PAWN] & _moveArray[color == WHITE and BPAWN or PAWN][cord]

    rayto = fromToRay[cord]
    blocker = board.blocker

    # Bishops and Queens
    if board.variant in ASEAN_VARIANTS:
        bishopMoves = _moveArray[ASEAN_BBISHOP if color == WHITE else ASEAN_WBISHOP]
        bits |= pieces[BISHOP] & bishopMoves[cord]

        bits |= pieces[QUEEN] & _moveArray[ASEAN_QUEEN][cord]
    else:
        bitboard = (pieces[BISHOP] | pieces[QUEEN] | pieces[HAWK]) & _moveArray[BISHOP][
            cord
        ]
        # inlined iterBits()
        # check whether or not there is a piece blocking the path in diagonal
        while bitboard:
            bit = bitboard & -bitboard
            c = lsb[bit]
            ray = rayto[c]
            if ray and not clearBit(ray & blocker, c):
                bits |= bitPosArray[c]
            bitboard -= bit

    # Rooks and queens
    if board.variant in ASEAN_VARIANTS:
        bitboard = pieces[ROOK] & _moveArray[ROOK][cord]
    else:
        bitboard = (pieces[ROOK] | pieces[QUEEN] | pieces[ELEPHANT]) & _moveArray[ROOK][
            cord
        ]
    # inlined iterBits()

    # check whether or not there is a piece blocking the path in straight line
    while bitboard:
        bit = bitboard & -bitboard
        c = lsb[bit]
        ray = rayto[c]
        if ray and not clearBit(ray & blocker, c):
            bits |= bitPosArray[c]
        bitboard -= bit

    return bits


def pinnedOnKing(board, cord, color):
    # Determine if the piece on cord is pinned against its colors king.
    # In chess, a pin is a situation in which a piece is forced to stay put
    # because moving it would expose a more valuable piece behind it to
    # capture.
    # Caveat: pinnedOnKing should only be called by genCheckEvasions().

    kingCord = board.kings[color]

    dir = directions[kingCord][cord]
    if dir == -1:
        return False

    opcolor = 1 - color
    blocker = board.blocker

    #  Path from piece to king is blocked, so no pin
    if clearBit(fromToRay[kingCord][cord], cord) & blocker:
        return False

    b = (rays[kingCord][dir] ^ fromToRay[kingCord][cord]) & blocker
    if not b:
        return False

    cord1 = cord > kingCord and firstBit(b) or lastBit(b)

    #  If diagonal
    if board.variant in ASEAN_VARIANTS:
        pass
    else:
        if dir <= 3 and bitPosArray[cord1] & (
            board.boards[opcolor][QUEEN]
            | board.boards[opcolor][BISHOP]
            | board.boards[opcolor][HAWK]
        ):
            return True

    #  Rank / file
    if board.variant in ASEAN_VARIANTS:
        if dir >= 4 and bitPosArray[cord1] & board.boards[opcolor][ROOK]:
            return True
    else:
        if dir >= 4 and bitPosArray[cord1] & (
            board.boards[opcolor][QUEEN]
            | board.boards[opcolor][ROOK]
            | board.boards[opcolor][ELEPHANT]
        ):
            return True

    return False


def staticExchangeEvaluate(board, moveOrTcord, color=None):
    """The GnuChess Static Exchange Evaluator (or SEE for short).
    First determine the target square.  Create a bitboard of all squares
    attacking the target square for both sides.  Using these 2 bitboards,
    we take turn making captures from smallest piece to largest piece.
    When a sliding piece makes a capture, we check behind it to see if
    another attacker piece has been exposed.  If so, add this to the bitboard
    as well.  When performing the "captures", we stop if one side is ahead
    and doesn't need to capture, a form of pseudo-minimaxing."""

    #
    # Notice: If you use the tcord version, the color is the color attacked, and
    #         the color to witch the score is relative.
    #

    swaplist = [0]

    LASTPIECE = ELEPHANT if board.variant == SCHESS else KING

    if color is None:
        move = moveOrTcord
        flag = move >> 12
        fcord = (move >> 6) & 63
        tcord = move & 63

        color = board.friends[BLACK] & bitPosArray[fcord] and BLACK or WHITE
        opcolor = 1 - color
        boards = board.boards[color]
        opboards = board.boards[opcolor]

        ours = getAttacks(board, tcord, color)
        ours = clearBit(ours, fcord)
        theirs = getAttacks(board, tcord, opcolor)

        if xray[board.arBoard[fcord]]:
            ours, theirs = addXrayPiece(board, tcord, fcord, color, ours, theirs)

        from pychess.Variants import variants

        PROMOTIONS = variants[board.variant].PROMOTIONS
        if flag in PROMOTIONS:
            swaplist = [PIECE_VALUES[flag - 3] - PAWN_VALUE]
            lastval = -PIECE_VALUES[flag - 3]
        else:
            if flag == ENPASSANT:
                swaplist = [PAWN_VALUE]
            else:
                swaplist = [PIECE_VALUES[board.arBoard[tcord]]]
            lastval = -PIECE_VALUES[board.arBoard[fcord]]

    else:
        tcord = moveOrTcord
        opcolor = 1 - color
        boards = board.boards[color]
        opboards = board.boards[opcolor]

        ours = getAttacks(board, tcord, color)
        theirs = getAttacks(board, tcord, opcolor)

        lastval = -PIECE_VALUES[board.arBoard[tcord]]

    while theirs:
        for piece in range(PAWN, LASTPIECE + 1):
            r = theirs & opboards[piece]
            if r:
                cord = firstBit(r)
                theirs = clearBit(theirs, cord)
                if xray[piece]:
                    ours, theirs = addXrayPiece(board, tcord, cord, color, ours, theirs)
                swaplist.append(swaplist[-1] + lastval)
                lastval = PIECE_VALUES[piece]
                break

        if not ours:
            break

        for piece in range(PAWN, LASTPIECE + 1):
            r = ours & boards[piece]
            if r:
                cord = firstBit(r)
                ours = clearBit(ours, cord)
                if xray[piece]:
                    ours, theirs = addXrayPiece(board, tcord, cord, color, ours, theirs)
                swaplist.append(swaplist[-1] + lastval)
                lastval = -PIECE_VALUES[piece]
                break

    #  At this stage, we have the swap scores in a list.  We just need to
    #  mini-max the scores from the bottom up to the top of the list.

    for n in range(len(swaplist) - 1, 0, -1):
        if n & 1:
            if swaplist[n] <= swaplist[n - 1]:
                swaplist[n - 1] = swaplist[n]
        else:
            if swaplist[n] >= swaplist[n - 1]:
                swaplist[n - 1] = swaplist[n]

    return swaplist[0]


xray = (False, True, False, True, True, True, False, True, True)


def addXrayPiece(board, tcord, fcord, color, ours, theirs):
    """This is used by swapOff.
    The purpose of this routine is to find a piece which attack through
    another piece (e.g. two rooks, Q+B, B+P, etc.) Color is the side attacking
    the square where the swapping is to be done."""

    dir = directions[tcord][fcord]
    a = rays[fcord][dir] & board.blocker
    if not a:
        return ours, theirs

    if tcord < fcord:
        ncord = firstBit(a)
    else:
        ncord = lastBit(a)

    piece = board.arBoard[ncord]
    if board.variant in ASEAN_VARIANTS:
        cond = piece == ROOK and dir > 3
    else:
        cond = piece == (
            QUEEN
            or (piece == ROOK and dir > 3)
            or (piece == ELEPHANT and dir > 3)
            or (piece == BISHOP and dir < 4)
            or (piece == HAWK and dir < 4)
        )
    if cond:
        bit = bitPosArray[ncord]
        if bit & board.friends[color]:
            ours |= bit
        else:
            theirs |= bit

    return ours, theirs


def defends(board, fcord, tcord):
    """Could fcord attack tcord if the piece on tcord wasn't on the team of
    fcord?
    Doesn't test check."""

    # Work on a board copy, as we are going to change some stuff
    board = board.clone()

    if board.friends[WHITE] & bitPosArray[fcord]:
        color = WHITE
    else:
        color = BLACK
    opcolor = 1 - color

    boards = board.boards[color]
    opboards = board.boards[opcolor]

    # To see if we now defend the piece, we have to "give" it to the other team
    piece = board.arBoard[tcord]

    # backup = boards[piece]
    # opbackup = opboards[piece]

    boards[piece] &= notBitPosArray[tcord]
    opboards[piece] |= bitPosArray[tcord]
    board.friends[color] &= notBitPosArray[tcord]
    board.friends[opcolor] |= bitPosArray[tcord]

    # Can we "attack" the piece now?
    backupColor = board.color
    board.setColor(color)
    from .lmovegen import newMove
    from .validator import validateMove

    islegal = validateMove(board, newMove(fcord, tcord))
    board.setColor(backupColor)

    # We don't need to set the board back, as we work on a copy
    # boards[piece] = backup
    # opboards[piece] = opbackup
    # board.friends[color] |= bitPosArray[tcord]
    # board.friends[opcolor] &= notBitPosArray[tcord]

    return islegal



"""Wildcastle Chess variant"""

from pychess.Utils.const import WILDCASTLECHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board

WILDCASTLESTART = "rnbkqbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"


class WildcastleBoard(Board):
    variant = WILDCASTLECHESS
    __desc__ = _(
        "xboard wildcastle http://home.hccnet.nl/h.g.muller/engine-intf.html#8\n"
        + "FICS wild/0: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "* White has the typical set-up at the start.\n"
        + "* Black's pieces are the same, except that the King and Queen are reversed,\n"
        + "* so they are not on the same files as White's King and Queen.\n"
        + "* Castling is done similarly to normal chess:\n"
        + "* o-o-o indicates long castling and o-o short castling."
    )
    name = _("Wildcastle")
    cecp_name = "wildcastle"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=WILDCASTLESTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import asyncio
import os
from io import StringIO

from gi.repository import GLib, Gtk, GObject

from pychess.ic import (
    IC_POS_EXAMINATING,
    IC_POS_OBSERVING_EXAMINATION,
    get_infobarmessage_content,
    get_infobarmessage_content2,
    TITLES,
)
from pychess.ic.ICGameModel import ICGameModel
from pychess.perspectives.fics.FicsHome import UserInfoSection
from pychess.perspectives.fics.SeekChallenge import SeekChallengeSection
from pychess.System import conf, uistuff
from pychess.System.prefix import addUserConfigPrefix
from pychess.System.Log import log
from pychess.widgets import new_notebook
from pychess.widgets.InfoBar import (
    InfoBarMessage,
    InfoBarNotebook,
    InfoBarMessageButton,
)
from pychess.widgets.pydock.PyDockTop import PyDockTop
from pychess.widgets.pydock import SOUTH, WEST, CENTER

from pychess.Utils.const import LOCAL, WHITE, BLACK, REMOTE
from pychess.Utils.TimeModel import TimeModel
from pychess.Players.ICPlayer import ICPlayer
from pychess.Players.Human import Human
from pychess.Savers import pgn, fen
from pychess.perspectives import Perspective, perspective_manager, panel_name


if not hasattr(Gtk.TreeModelFilter, "new_with_model"):
    # Fix #1811: TreeModelFilter.sort_new_with_model() is missing on some Gtk versions
    # due to API changes. Let's keep compatibility with older versions.

    def sort_new_with_model(self):
        super_object = super(Gtk.TreeModel, self)
        if hasattr(super_object, "sort_new_with_model"):
            return super_object.sort_new_with_model()
        return Gtk.TreeModelSort.new_with_model(self)

    @classmethod
    def new_with_model(self, child_model):
        return Gtk.TreeModel.sort_new_with_model(child_model)

    Gtk.TreeModel.sort_new_with_model = sort_new_with_model
    Gtk.TreeModelFilter.new_with_model = new_with_model


class PlayerNotificationMessage(InfoBarMessage):
    def __init__(self, message_type, content, callback, player, text):
        InfoBarMessage.__init__(self, message_type, content, callback)
        self.player = player
        self.text = text


class FICS(GObject.GObject, Perspective):
    __gsignals__ = {
        "logout": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "autoLogout": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        log.debug("FICS.__init__: starting")
        GObject.GObject.__init__(self)
        Perspective.__init__(self, "fics", _("ICS"))
        self.dockLocation = addUserConfigPrefix("pydock-fics.xml")
        self.first_run = True

    def create_toolbuttons(self):
        def on_logoff_clicked(button):
            self.emit("logout")
            self.close()

        self.logoff_button = Gtk.ToolButton.new_from_stock(Gtk.STOCK_DISCONNECT)
        self.logoff_button.set_tooltip_text(_("Log Off"))
        self.logoff_button.set_label("logoff")
        self.logoff_button.connect("clicked", on_logoff_clicked)

        def on_minute_1_clicked(button):
            self.connection.client.run_command("1-minute")

        def on_minute_3_clicked(button):
            self.connection.client.run_command("3-minute")

        def on_minute_5_clicked(button):
            self.connection.client.run_command("5-minute")

        def on_minute_15_clicked(button):
            self.connection.client.run_command("15-minute")

        def on_minute_25_clicked(button):
            self.connection.client.run_command("25-minute")

        def on_chess960_clicked(button):
            self.connection.client.run_command("chess960")

        self.minute_1_button = Gtk.ToggleToolButton()
        self.minute_1_button.set_label("1")
        self.minute_1_button.set_tooltip_text(_("New game from 1-minute playing pool"))
        self.minute_1_button.connect("clicked", on_minute_1_clicked)

        self.minute_3_button = Gtk.ToggleToolButton()
        self.minute_3_button.set_label("3")
        self.minute_3_button.set_tooltip_text(_("New game from 3-minute playing pool"))
        self.minute_3_button.connect("clicked", on_minute_3_clicked)

        self.minute_5_button = Gtk.ToggleToolButton()
        self.minute_5_button.set_label("5")
        self.minute_5_button.set_tooltip_text(_("New game from 5-minute playing pool"))
        self.minute_5_button.connect("clicked", on_minute_5_clicked)

        self.minute_15_button = Gtk.ToggleToolButton()
        self.minute_15_button.set_label("15")
        self.minute_15_button.set_tooltip_text(
            _("New game from 15-minute playing pool")
        )
        self.minute_15_button.connect("clicked", on_minute_15_clicked)

        self.minute_25_button = Gtk.ToggleToolButton()
        self.minute_25_button.set_label("25")
        self.minute_25_button.set_tooltip_text(
            _("New game from 25-minute playing pool")
        )
        self.minute_25_button.connect("clicked", on_minute_25_clicked)

        self.chess960_button = Gtk.ToggleToolButton()
        self.chess960_button.set_label("960")
        self.chess960_button.set_tooltip_text(_("New game from Chess960 playing pool"))
        self.chess960_button.connect("clicked", on_chess960_clicked)

    def init_layout(self):
        perspective_widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        perspective_manager.set_perspective_widget("fics", perspective_widget)

        self.infobar = InfoBarNotebook("fics_lounge_infobar")
        self.infobar.hide()
        perspective_widget.pack_start(self.infobar, False, False, 0)

        self.dock = PyDockTop("fics", self)
        align = Gtk.Alignment()
        align.show()
        align.add(self.dock)
        self.dock.show()
        perspective_widget.pack_start(align, True, True, 0)

        self.notebooks = {"ficshome": new_notebook()}
        self.main_notebook = self.notebooks["ficshome"]
        for panel in self.sidePanels:
            self.notebooks[panel_name(panel.__name__)] = new_notebook(
                panel_name(panel.__name__)
            )

        self.docks["ficshome"] = (
            Gtk.Label(label="ficshome"),
            self.notebooks["ficshome"],
            None,
        )
        for panel in self.sidePanels:
            self.docks[panel_name(panel.__name__)][1] = self.notebooks[
                panel_name(panel.__name__)
            ]

        self.load_from_xml()

        # Default layout of side panels
        first_time_layout = False
        if not os.path.isfile(self.dockLocation):
            first_time_layout = True
            leaf = self.dock.dock(
                self.docks["ficshome"][1], CENTER, self.docks["ficshome"][0], "ficshome"
            )
            leaf.setDockable(False)

            console_leaf = leaf.dock(
                self.docks["ConsolePanel"][1],
                SOUTH,
                self.docks["ConsolePanel"][0],
                "ConsolePanel",
            )
            console_leaf.dock(
                self.docks["NewsPanel"][1],
                CENTER,
                self.docks["NewsPanel"][0],
                "NewsPanel",
            )

            seek_leaf = leaf.dock(
                self.docks["SeekListPanel"][1],
                WEST,
                self.docks["SeekListPanel"][0],
                "SeekListPanel",
            )
            seek_leaf.dock(
                self.docks["SeekGraphPanel"][1],
                CENTER,
                self.docks["SeekGraphPanel"][0],
                "SeekGraphPanel",
            )
            seek_leaf.dock(
                self.docks["PlayerListPanel"][1],
                CENTER,
                self.docks["PlayerListPanel"][0],
                "PlayerListPanel",
            )
            seek_leaf.dock(
                self.docks["GameListPanel"][1],
                CENTER,
                self.docks["GameListPanel"][0],
                "GameListPanel",
            )
            seek_leaf.dock(
                self.docks["ArchiveListPanel"][1],
                CENTER,
                self.docks["ArchiveListPanel"][0],
                "ArchiveListPanel",
            )

            leaf = leaf.dock(
                self.docks["ChatPanel"][1],
                SOUTH,
                self.docks["ChatPanel"][0],
                "ChatPanel",
            )
            # leaf.dock(self.docks["LecturesPanel"][1], CENTER, self.docks["LecturesPanel"][0], "LecturesPanel")

        def unrealize(dock):
            dock.saveToXML(self.dockLocation)
            dock._del()

        self.dock.connect("unrealize", unrealize)

        self.dock.show_all()
        perspective_widget.show_all()

        perspective_manager.set_perspective_menuitems(
            "fics", self.menuitems, default=first_time_layout
        )

        log.debug("FICS.__init__: finished")

    def open_lounge(self, connection, helperconn, host):
        if self.first_run:
            self.init_layout()

        self.connection = connection
        self.helperconn = helperconn
        self.host = host

        self.finger_sent = False
        self.messages = []
        self.players = []
        self.game_cids = {}

        self.widgets = uistuff.GladeWidgets("fics_lounge.glade")
        self.widgets["fics_lounge"].hide()

        fics_home = self.widgets["fics_home"]
        self.widgets["fics_lounge_content_hbox"].remove(fics_home)

        self.archive_list = self.widgets["archiveListContent"]
        self.widgets["fics_panels_notebook"].remove(self.archive_list)

        self.games_list = self.widgets["gamesListContent"]
        self.widgets["fics_panels_notebook"].remove(self.games_list)

        self.news_list = self.widgets["news"]
        self.widgets["fics_home"].remove(self.news_list)

        self.players_list = self.widgets["playersListContent"]
        self.widgets["fics_panels_notebook"].remove(self.players_list)

        self.seek_graph = self.widgets["seekGraphContent"]
        self.widgets["fics_panels_notebook"].remove(self.seek_graph)

        self.seek_list = self.widgets["seekListContent"]
        self.widgets["fics_panels_notebook"].remove(self.seek_list)

        self.seek_challenge = SeekChallengeSection(self)

        def on_autoLogout(alm):
            self.emit("autoLogout")
            self.close()

        self.connection.alm.connect("logOut", on_autoLogout)
        self.connection.connect("disconnected", lambda connection: self.close())
        self.connection.connect("error", self.on_connection_error)
        if self.connection.isRegistred():
            numtimes = conf.get("numberOfTimesLoggedInAsRegisteredUser") + 1
            conf.set("numberOfTimesLoggedInAsRegisteredUser", numtimes)
        self.connection.em.connect(
            "onCommandNotFound",
            lambda em, cmd: log.error("Fics answered '%s': Command not found" % cmd),
        )
        self.connection.bm.connect("playGameCreated", self.onPlayGameCreated)
        self.connection.bm.connect("obsGameCreated", self.onObserveGameCreated)
        self.connection.bm.connect("exGameCreated", self.onObserveGameCreated)
        self.connection.fm.connect("fingeringFinished", self.onFinger)
        # the rest of these relay server messages to the lounge infobar
        self.connection.bm.connect("tooManySeeks", self.tooManySeeks)
        self.connection.bm.connect("nonoWhileExamine", self.nonoWhileExamine)
        self.connection.bm.connect("matchDeclined", self.matchDeclined)
        self.connection.bm.connect("player_on_censor", self.player_on_censor)
        self.connection.bm.connect("player_on_noplay", self.player_on_noplay)
        self.connection.bm.connect("req_not_fit_formula", self.req_not_fit_formula)
        self.connection.glm.connect("seek-updated", self.on_seek_updated)
        self.connection.glm.connect("our-seeks-removed", self.our_seeks_removed)
        self.connection.cm.connect("arrivalNotification", self.onArrivalNotification)
        self.connection.cm.connect("departedNotification", self.onDepartedNotification)

        def get_top_games():
            if perspective_manager.current_perspective == self:
                self.connection.client.run_command("games *19")
            return True

        if self.connection.ICC:
            self.event_id = GLib.timeout_add_seconds(5, get_top_games)

        for user in self.connection.notify_users:
            user = self.connection.players.get(user)
            self.user_from_notify_list_is_present(user)

        self.userinfo = UserInfoSection(self.widgets, self.connection, self.host, self)
        if not self.first_run:
            self.notebooks["ficshome"].remove_page(-1)
        self.notebooks["ficshome"].append_page(fics_home)

        self.panels = [
            panel.Sidepanel().load(self.widgets, self.connection, self)
            for panel in self.sidePanels
        ]

        for panel, instance in zip(self.sidePanels, self.panels):
            if not self.first_run:
                self.notebooks[panel_name(panel.__name__)].remove_page(-1)
            self.notebooks[panel_name(panel.__name__)].append_page(instance)
            instance.show()

        tool_buttons = [
            self.logoff_button,
        ]
        self.quick_seek_buttons = []
        if self.connection.ICC:
            self.quick_seek_buttons = [
                self.minute_1_button,
                self.minute_3_button,
                self.minute_5_button,
                self.minute_15_button,
                self.minute_25_button,
                self.chess960_button,
            ]
            tool_buttons += self.quick_seek_buttons
        perspective_manager.set_perspective_toolbuttons("fics", tool_buttons)

        if self.first_run:
            self.first_run = False

        # After all panel is set up we can push initial messages out
        self.connection.com.onConsoleMessage("", self.connection.ini_messages)

    def show(self):
        perspective_manager.activate_perspective("fics")

    def present(self):
        perspective_manager.activate_perspective("fics")

    def on_connection_error(self, connection, error):
        log.warning("FICS.on_connection_error: %s" % repr(error))
        self.close()

    def close(self):
        try:
            self.widgets = None
        except TypeError:
            pass
        except AttributeError:
            pass
        perspective_manager.disable_perspective("fics")

    def onPlayGameCreated(self, bm, ficsgame):
        log.debug("FICS.onPlayGameCreated: %s" % ficsgame)

        for message in self.messages:
            message.dismiss()
        del self.messages[:]

        if self.connection.ICC:
            for button in self.quick_seek_buttons:
                button.set_active(False)

        timemodel = TimeModel(ficsgame.minutes * 60, ficsgame.inc)

        gamemodel = ICGameModel(self.connection, ficsgame, timemodel)
        gamemodel.connect("game_started", self.onGameModelStarted, ficsgame)

        wplayer, bplayer = ficsgame.wplayer, ficsgame.bplayer

        # We start
        if wplayer.name.lower() == self.connection.getUsername().lower():
            player0tup = (
                LOCAL,
                Human,
                (
                    WHITE,
                    wplayer.long_name(),
                    wplayer.name,
                    wplayer.getRatingForCurrentGame(),
                ),
                wplayer.long_name(),
            )
            player1tup = (
                REMOTE,
                ICPlayer,
                (
                    gamemodel,
                    bplayer.name,
                    ficsgame.gameno,
                    BLACK,
                    bplayer.long_name(),
                    bplayer.getRatingForCurrentGame(),
                ),
                bplayer.long_name(),
            )

        # She starts
        else:
            player1tup = (
                LOCAL,
                Human,
                (
                    BLACK,
                    bplayer.long_name(),
                    bplayer.name,
                    bplayer.getRatingForCurrentGame(),
                ),
                bplayer.long_name(),
            )
            player0tup = (
                REMOTE,
                ICPlayer,
                (
                    gamemodel,
                    wplayer.name,
                    ficsgame.gameno,
                    WHITE,
                    wplayer.long_name(),
                    wplayer.getRatingForCurrentGame(),
                ),
                wplayer.long_name(),
            )

        perspective = perspective_manager.get_perspective("games")
        if not ficsgame.board.fen:
            asyncio.create_task(
                perspective.generalStart(gamemodel, player0tup, player1tup)
            )
        else:
            asyncio.create_task(
                perspective.generalStart(
                    gamemodel,
                    player0tup,
                    player1tup,
                    (StringIO(ficsgame.board.fen), fen, 0, -1),
                )
            )

    def onGameModelStarted(self, gamemodel, ficsgame):
        self.connection.bm.onGameModelStarted(ficsgame.gameno)

    def onObserveGameCreated(self, bm, ficsgame):
        log.debug("FICS.onObserveGameCreated: %s" % ficsgame)

        timemodel = TimeModel(ficsgame.minutes * 60, ficsgame.inc)

        gamemodel = ICGameModel(self.connection, ficsgame, timemodel)
        gamemodel.connect("game_started", self.onGameModelStarted, ficsgame)

        # The players need to start listening for moves IN this method if they
        # want to be noticed of all moves the FICS server sends us from now on
        wplayer, bplayer = ficsgame.wplayer, ficsgame.bplayer

        player0tup = (
            REMOTE,
            ICPlayer,
            (
                gamemodel,
                wplayer.name,
                ficsgame.gameno,
                WHITE,
                wplayer.long_name(),
                wplayer.getRatingForCurrentGame(),
            ),
            wplayer.long_name(),
        )
        player1tup = (
            REMOTE,
            ICPlayer,
            (
                gamemodel,
                bplayer.name,
                ficsgame.gameno,
                BLACK,
                bplayer.long_name(),
                bplayer.getRatingForCurrentGame(),
            ),
            bplayer.long_name(),
        )

        perspective = perspective_manager.get_perspective("games")
        asyncio.create_task(
            perspective.generalStart(
                gamemodel,
                player0tup,
                player1tup,
                (StringIO(ficsgame.board.pgn), pgn, 0, -1),
            )
        )

        if ficsgame.relation == IC_POS_OBSERVING_EXAMINATION:
            if 1:  # int(self.connection.lvm.variablesBackup["kibitz"]) == 0:
                self.connection.cm.whisper(
                    _("You have to set kibitz on to see bot messages here.")
                )
            self.connection.fm.finger(bplayer.name)
            self.connection.fm.finger(wplayer.name)
        elif ficsgame.relation == IC_POS_EXAMINATING:
            gamemodel.examined = True
        if not self.connection.ICC:
            allob = "allob " + str(ficsgame.gameno)
            gamemodel.connection.client.run_command(allob)

    def onFinger(self, fm, finger):
        titles = finger.getTitles()
        if titles is not None:
            name = finger.getName()
            player = self.connection.players.get(name)
            for title in titles:
                player.titles.add(TITLES[title])

    def tooManySeeks(self, bm):
        label = Gtk.Label(
            label=_(
                "You may only have 3 outstanding seeks at the same time. If you want \
            to add a new seek you must clear your currently active seeks. Clear your seeks?"
            )
        )
        label.set_width_chars(80)
        label.props.xalign = 0
        label.set_line_wrap(True)

        def response_cb(infobar, response, message):
            if response == Gtk.ResponseType.YES:
                self.connection.client.run_command("unseek")
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.QUESTION, label, response_cb)
        message.add_button(InfoBarMessageButton(Gtk.STOCK_YES, Gtk.ResponseType.YES))
        message.add_button(InfoBarMessageButton(Gtk.STOCK_NO, Gtk.ResponseType.NO))
        self.messages.append(message)
        self.infobar.push_message(message)

    def nonoWhileExamine(self, bm):
        label = Gtk.Label(_("You can't touch this! You are examining a game."))

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, label, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def matchDeclined(self, bm, player):
        text = _(" has declined your offer for a match")
        content = get_infobarmessage_content(player, text)

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def player_on_censor(self, bm, player):
        text = _(" is censoring you")
        content = get_infobarmessage_content(player, text)

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def player_on_noplay(self, bm, player):
        text = _(" noplay listing you")
        content = get_infobarmessage_content(player, text)

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def req_not_fit_formula(self, bm, player, formula):
        content = get_infobarmessage_content2(
            player, _(" uses a formula not fitting your match request:"), formula
        )

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def on_seek_updated(self, glm, message_text):
        if "manual accept" in message_text:
            message_text.replace("to manual accept", _("to manual accept"))
        elif "automatic accept" in message_text:
            message_text.replace("to automatic accept", _("to automatic accept"))
        if "rating range now" in message_text:
            message_text.replace("rating range now", _("rating range now"))
        label = Gtk.Label(label=_("Seek updated") + ": " + message_text)

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, label, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def our_seeks_removed(self, glm):
        label = Gtk.Label(label=_("Your seeks have been removed"))

        def response_cb(infobar, response, message):
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.INFO, label, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def _connect_to_player_changes(self, player):
        player.connect("ratings_changed", self._replace_notification_message, player)
        player.connect(
            "notify::titles", self._replace_notification_message, None, player
        )

    def onArrivalNotification(self, cm, player):
        log.debug(
            "%s" % player,
            extra={"task": (self.connection.username, "onArrivalNotification")},
        )
        self._add_notification_message(
            player, _(" has arrived"), chat=True, replace=True
        )
        if player not in self.players:
            self.players.append(player)
            self._connect_to_player_changes(player)

    def onDepartedNotification(self, cm, player):
        self._add_notification_message(player, _(" has departed"), replace=True)

    def user_from_notify_list_is_present(self, player):
        self._add_notification_message(
            player, _(" is present"), chat=True, replace=True
        )
        if player not in self.players:
            self.players.append(player)
            self._connect_to_player_changes(player)

    def _add_notification_message(self, player, text, chat=False, replace=False):
        if replace:
            for message in self.messages:
                if (
                    isinstance(message, PlayerNotificationMessage)
                    and message.player == player
                ):
                    message.dismiss()

        content = get_infobarmessage_content(player, text)

        def response_cb(infobar, response, message):
            if response == 1:
                if player is None:
                    return
                self.chat.openChatWithPlayer(player.name)
            if response == 2:
                if player is None:
                    return
                self.connection.client.run_command("follow %s" % player.name)
            message.dismiss()
            #             self.messages.remove(message)
            return False

        message = PlayerNotificationMessage(
            Gtk.MessageType.INFO, content, response_cb, player, text
        )
        if chat:
            message.add_button(InfoBarMessageButton(_("Chat"), 1))
            message.add_button(InfoBarMessageButton(_("Follow"), 2))
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages.append(message)
        self.infobar.push_message(message)

    def _replace_notification_message(self, obj, prop, rating_type, player):
        log.debug(
            f"{repr(obj)} {player}",
            extra={"task": (self.connection.username, "_replace_notification_message")},
        )
        for message in self.messages:
            if (
                isinstance(message, PlayerNotificationMessage)
                and message.player == player
            ):
                message.update_content(get_infobarmessage_content(player, message.text))
        return False
